"""
Tests for Pyro Flame.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import sys
import unittest
import Pyro4.utils.flame
import Pyro4.utils.flameserver
import Pyro4.core
import Pyro4.errors
import Pyro4.constants
from Pyro4.configuration import config
from testsupport import *


class FlameDisabledTests(unittest.TestCase):
    def testFlameDisabled(self):
        with Pyro4.core.Daemon() as d:
            with self.assertRaises(Pyro4.errors.SecurityError) as ex:
                Pyro4.utils.flame.start(d)
            self.assertIn("disabled", str(ex.exception))   # default should be disabled even when pickle is activated

    def testRequirePickleExclusively(self):
        with Pyro4.core.Daemon() as d:
            config.FLAME_ENABLED = True
            sers = config.SERIALIZERS_ACCEPTED
            config.SERIALIZERS_ACCEPTED = {"json", "serpent", "marshal"}
            self.assertRaises(Pyro4.errors.SerializeError, Pyro4.utils.flame.start, d)  # require pickle
            config.SERIALIZERS_ACCEPTED.add("pickle")
            self.assertRaises(Pyro4.errors.SerializeError, Pyro4.utils.flame.start, d)  # require pickle exclusively
            config.SERIALIZERS_ACCEPTED = {"pickle"}
            Pyro4.utils.flame.start(d)
            config.SERIALIZERS_ACCEPTED = sers
            config.FLAME_ENABLED = False


class FlameTests(unittest.TestCase):
    def setUp(self):
        config.FLAME_ENABLED = True
        self._serializers = config.SERIALIZERS_ACCEPTED
        config.SERIALIZERS_ACCEPTED = {"pickle"}

    def tearDown(self):
        config.FLAME_ENABLED = False
        config.SERIALIZERS_ACCEPTED = self._serializers

    def testExposed(self):
        e = Pyro4.utils.flame.Flame()
        self.assertTrue(hasattr(e, "_pyroExposed"))

    def testCreateModule(self):
        module = Pyro4.utils.flame.createModule("testmodule", "def x(y): return y*y")
        self.assertEqual(9, module.x(3))
        module = Pyro4.utils.flame.createModule("testmodule2.submodule.subsub", "def x(y): return y*y")
        self.assertEqual(9, module.x(3))
        import testmodule2.submodule.subsub
        self.assertEqual(9, testmodule2.submodule.subsub.x(3))

    def testCreateModuleNamespace(self):
        namespace = {}
        Pyro4.utils.flame.createModule("testmodule2.submodule.subsub", "def x(y): return y*y", namespace=namespace)
        self.assertEqual(9, namespace["testmodule2"].submodule.subsub.x(3))

    def testExecFunction(self):
        namespace = {}
        Pyro4.utils.flame.exec_function("foobar=5+6", "<foo>", namespace)
        self.assertEqual(11, namespace["foobar"])

    def testExecFunctionNewlines(self):
        namespace = {}
        Pyro4.utils.flame.exec_function("if True:\r\n  foobar=5+6\r\n   ", "<foo>", namespace)
        self.assertEqual(11, namespace["foobar"])

    def testFlameModule(self):
        with Pyro4.core.Daemon() as d:
            Pyro4.utils.flame.start(d)
            flameserver = d.objectsById[Pyro4.constants.FLAME_NAME]
            with Pyro4.utils.flame.FlameModule(flameserver, "sys") as m:
                self.assertIn("module 'sys' at", str(m))
                self.assertIsInstance(m.exc_info, Pyro4.core._RemoteMethod)

    def testFlameBuiltin(self):
        with Pyro4.core.Daemon() as d:
            Pyro4.utils.flame.start(d)
            flameserver = d.objectsById[Pyro4.constants.FLAME_NAME]
            with Pyro4.utils.flame.FlameBuiltin(flameserver, "max") as builtin:
                self.assertTrue(hasattr(builtin, "__call__"))
                self.assertIn("builtin 'max' at", str(builtin))

    def testFlameserverMain(self):
        oldstdout = sys.stdout
        oldstderr = sys.stderr
        try:
            sys.stdout = StringIO()
            sys.stderr = StringIO()
            self.assertRaises(SystemExit, Pyro4.utils.flameserver.main, ["--invalidarg"])
            self.assertTrue("no such option" in sys.stderr.getvalue())
            sys.stderr.truncate(0)
            sys.stdout.truncate(0)
            self.assertRaises(SystemExit, Pyro4.utils.flameserver.main, ["-h"])
            self.assertTrue("show this help message" in sys.stdout.getvalue())
        finally:
            sys.stdout = oldstdout
            sys.stderr = oldstderr


if __name__ == "__main__":
    # import sys; sys.argv = ['', 'Test.testName']
    unittest.main()



import time
import Pyro4


@Pyro4.expose
class Thingy(object):
    def multiply(self, a, b):
        return a * b

    def add(self, a, b):
        return a + b

    def divide(self, a, b):
        return a // b

    def error(self):
        return 1 // 0

    def delay(self, seconds):
        time.sleep(seconds)
        return seconds

    def printmessage(self, message):
        print(message)
        return 0


Pyro4.Daemon.serveSimple({
    Thingy: "example.batched"
}, ns=False)



# this only works on Linux
# it uses the abstract namespace socket feature.

from __future__ import print_function
import Pyro4


@Pyro4.expose
class Thingy(object):
    def message(self, arg):
        print("Message received:", arg)
        return "Roger!"


with Pyro4.Daemon(unixsocket="\0example_unix.sock") as d:   # notice the 0-byte at the start
    uri = d.register(Thingy, "example.unixsock")
    print("Server running, uri=", uri)
    string_uri = str(uri)
    print("Actually, the uri contains a 0-byte, make sure you copy the part between the quotes to the client:")
    print(repr(string_uri))
    d.requestLoop()



from __future__ import print_function
import time

import Pyro4


@Pyro4.expose
@Pyro4.behavior(instance_mode="single")
class RemoteObject(object):
    def __init__(self):
        self.amount = 0

    def method(self, arg):
        return " ~~this is the remote result~~ "

    def work(self):
        print("work... %d" % self.amount)
        time.sleep(0.5)
        self.amount += 1

    def reset_work(self):
        self.amount = 0

    def get_work_done(self):
        return self.amount


Pyro4.Daemon.serveSimple({
    RemoteObject: "example.proxysharing"
})



# this example forks() and thus won't work on Windows.

from __future__ import print_function
import os
import signal
import socket
import Pyro4

# it's okay to use Pickle because we trust our own processes, and it is efficient
Pyro4.config.SERIALIZER = "pickle"
Pyro4.config.SERIALIZERS_ACCEPTED |= {"pickle"}


# create our own socket pair (server-client sockets that are already connected)
sock1, sock2 = socket.socketpair()

pid = os.fork()

if pid == 0:
    # we are the child process, we host the daemon.

    class Echo(object):
        @Pyro4.expose
        def echo(self, message):
            print("server got message: ", message)
            return "thank you"

    # create a daemon with some Pyro objectrunning on our custom server socket
    daemon = Pyro4.Daemon(connected_socket=sock1)
    daemon.register(Echo, "echo")
    print("Process PID={:d}: Pyro daemon running on {:s}\n".format(os.getpid(), daemon.locationStr))
    daemon.requestLoop()

else:
    # we are the parent process, we create a Pyro client proxy
    print("Process PID={:d}: Pyro client.\n".format(os.getpid()))

    # create a client running on the client socket
    with Pyro4.Proxy("echo", connected_socket=sock2) as p:
        reply = p.echo("hello!")
        print("client got reply:", reply)
        reply = p.echo("hello again!")
        print("client got reply:", reply)
    with Pyro4.Proxy("echo", connected_socket=sock2) as p:
        reply = p.echo("hello2!")
        print("client got reply:", reply)
        reply = p.echo("hello2 again!")
        print("client got reply:", reply)

    os.kill(pid, signal.SIGTERM)
    os.waitpid(pid, 0)
    print("\nThe end.")



from __future__ import print_function
import sys
import array


class Wall(object):
    """an obstructing static wall"""

    def __init__(self, position):
        self.x, self.y = position

    def __getstate__(self):
        return self.x, self.y

    def __setstate__(self, state):
        self.x, self.y = state


class Robot(object):
    """represents a robot moving on a grid."""

    def __init__(self, name, grid_dimensions, position, direction=(0, 0), strength=5):
        self.name = name
        self.x, self.y = position
        self.dx, self.dy = direction
        self.gridw, self.gridh = grid_dimensions
        self.strength = strength

    def __str__(self):
        return "ROBOT '%s'; pos(%d,%d); dir(%d,%d); strength %d" % (self.name, self.x, self.y, self.dx, self.dy, self.strength)

    @staticmethod
    def dict_to_robot(classname, data):
        assert classname == "robot.Robot"
        return Robot(data["name"], data["grid_dimensions"], data["position"], data["direction"], data["strength"])

    @staticmethod
    def robot_to_dict(robot):
        return {
            "__class__": "robot.Robot",
            "name": robot.name,
            "grid_dimensions": (robot.gridw, robot.gridh),
            "position": (robot.x, robot.y),
            "direction": (robot.dx, robot.dy),
            "strength": robot.strength,
        }

    def move(self, world=None):
        # minmax to avoid moving off the sides
        x = min(self.gridw - 1, max(0, self.x + self.dx))
        y = min(self.gridh - 1, max(0, self.y + self.dy))
        if x == self.x and y == self.y:
            return
        if world and self.__process_collision(x, y, world):
            return
        self.x, self.y = x, y

    def __process_collision(self, newx, newy, world):
        other = world.collides(newx, newy)
        if not other:
            return False  # we didn't hit anything
        self.dx, self.dy = 0, 0  # come to a standstill when we hit something
        if isinstance(other, Wall):
            self.strength -= 1  # hit wall, decrease our strength
            if self.strength <= 0:
                print("[server] %s killed himself!" % self.name)
                world.remove(self)
                self.died(None, world)
        else:
            other.strength -= 1  # hit other robot, decrease other robot's strength
            self.collision(other)
            if other.strength <= 0:
                world.remove(other)
                other.died(self, world)
                print("[server] %s killed %s!" % (self.name, other.name))
        return True

    def killed(self, victim, world):
        """you can override this to react on kills"""
        pass

    def collision(self, other):
        """you can override this to react on collisions between bots"""
        pass

    def emote(self, text):
        """you can override this"""
        print("[server] %s says: '%s'" % (self.name, text))


class World(object):
    """the world the robots move in (Cartesian grid)"""

    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.all = []
        self.robots = []

    def add_wall(self, wall):
        self.all.append(wall)

    def add_robot(self, bot):
        self.all.append(bot)
        self.robots.append(bot)

    def collides(self, x, y):
        for obj in self.all:
            if obj.x == x and obj.y == y:
                return obj
        return None

    def remove(self, obj):
        self.all.remove(obj)
        self.robots.remove(obj)

    def dump(self):
        line = ' ' * self.width
        if sys.version_info >= (3, 0):
            line = bytes(line, "ASCII")
        grid = [array.array('b', line) for y in range(self.height)]
        for obj in self.all:
            grid[obj.y][obj.x] = ord('R') if isinstance(obj, Robot) else ord('#')
        return grid

    def __getstate__(self):
        return self.width, self.height, self.all, self.robots

    def __setstate__(self, state):
        self.width, self.height, self.all, self.robots = state



from __future__ import print_function, division
import time
import Pyro4


if Pyro4.config.ITER_STREAMING:
    print("Note: iter-streaming has been enabled in the Pyro config.")
else:
    print("Note: iter-streaming has not been enabled in the Pyro config (PYRO_ITER_STREAMING).")


@Pyro4.expose
class Streamer(object):
    def list(self):
        return [1, 2, 3, 4, 5, 6, 7, 8, 9]

    def iterator(self):
        return iter([1, 2, 3, 4, 5, 6, 7, 8, 9])

    def generator(self):
        i = 1
        while i < 10:
            yield i
            i += 1

    def slow_generator(self):
        i = 1
        while i < 10:
            time.sleep(0.5)
            yield i
            i += 1

    def fibonacci(self):
        a, b = 0, 1
        while True:
            yield a
            a, b = b, a + b


Pyro4.Daemon.serveSimple({
        Streamer: "example.streamer"
    }, ns=False)



"""
Tests for a running Pyro server, with timeouts.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import unittest
import test_server


class ServerTestsThreadTimeout(test_server.ServerTestsThreadNoTimeout):
    SERVERTYPE = "thread"
    COMMTIMEOUT = 2.0

    def testServerParallelism(self):
        # this test is not suitable on a server with timeout set
        pass

    def testProxySharing(self):
        pass

    def testException(self):
        pass


class ServerTestsMultiplexTimeout(test_server.ServerTestsMultiplexNoTimeout):
    SERVERTYPE = "multiplex"
    COMMTIMEOUT = 2.0

    def testServerParallelism(self):
        # this test is not suitable on a server with timeout set
        pass

    def testProxySharing(self):
        pass

    def testException(self):
        pass


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



class Workitem(object):
    def __init__(self, itemId, data):
        print("Created workitem %s" % itemId)
        self.itemId = itemId
        self.data = data
        self.result = None
        self.processedBy = None

    def __str__(self):
        return "<Workitem id=%s>" % str(self.itemId)

    @staticmethod
    def from_dict(classname, d):
        """this method is used to deserialize a workitem from Pyro"""
        assert classname == "workitem.Workitem"
        w = Workitem(d["itemId"], d["data"])
        w.result = d["result"]
        w.processedBy = d["processedBy"]
        return w



from __future__ import print_function, division
import sys
import Pyro4


@Pyro4.expose
class Mandelbrot(object):
    maxiters = 500

    def calc_line(self, start, res_x, ii, dr, line_nr):
        line = ""
        z = start + complex(0, ii)
        for r in range(res_x):
            z += complex(dr, 0)
            iters = self.iterations(z)
            line += " " if iters >= self.maxiters else chr(iters % 64 + 32)
        return line_nr, line

    def calc_lines(self, start, res_x, dr, di, start_line_nr, num_lines):
        lines = []
        for i in range(num_lines):
            line = ""
            for r in range(res_x):
                z = start + complex(r*dr, i*di)
                iters = self.iterations(z)
                line += " " if iters >= self.maxiters else chr(iters % 64 + 32)
            lines.append((i+start_line_nr, line))
        return lines

    def iterations(self, z):
        c = z
        for n in range(self.maxiters):
            if abs(z) > 2:
                return n
            z = z*z + c
        return self.maxiters


@Pyro4.expose
class MandelbrotColorPixels(object):
    maxiters = 500
    def calc_photoimage_line(self, y, res_x, res_y):
        line = []
        for x in range(res_x):
            rgb = self.mandel_iterate(x, y, res_x, res_y)
            line.append(rgb)
        # tailored response for easy drawing into a tkinter PhotoImage:
        return y, "{"+" ".join("#%02x%02x%02x" % rgb for rgb in line)+"}"

    def mandel_iterate(self, x, y, res_x, res_y):
        zr = (x/res_x - 0.5) * 1 - 0.3
        zi = (y/res_y - 0.5) * 1 - 0.9
        zi *= res_y/res_x  # aspect correction
        z = complex(zr, zi)
        c = z
        iters = 0
        for iters in range(self.maxiters+1):
            if abs(z) > 2:
                break
            z = z*z + c
        if iters >= self.maxiters:
            return 0, 0, 0
        r = (iters+32) % 255
        g = iters % 255
        b = (iters+40) % 255
        return int(r), int(g), int(b)


if __name__ == "__main__":
    # spawn a Pyro daemon process
    # (can't use threads, because of the GIL)
    if len(sys.argv) != 2:
        raise SystemExit("give argument: server_id number")

    server_id = int(sys.argv[1])
    with Pyro4.Daemon() as d:
        with Pyro4.locateNS() as ns:
            mandel_server = d.register(Mandelbrot)
            mandel_color_server = d.register(MandelbrotColorPixels)
            ns.register("mandelbrot_"+str(server_id), mandel_server, safe=True, metadata={"class:mandelbrot_calc"})
            ns.register("mandelbrot_color_"+str(server_id), mandel_color_server, safe=True, metadata={"class:mandelbrot_calc_color"})
        print("Mandelbrot calculation server #{} ready.".format(server_id))
        d.requestLoop()



from __future__ import print_function
import sys
import time
import Pyro4
import bench

Pyro4.config.SERIALIZER = "marshal"

if sys.version_info < (3, 0):
    input = raw_input

uri = input("Uri of benchmark server? ").strip()
obj = Pyro4.core.Proxy(uri)
obj._pyroBind()
assert "oneway" in obj._pyroOneway   # make sure this method is indeed marked as @oneway


funcs = [
    lambda: obj.length('Irmen de Jong'),
    lambda: obj.timestwo(21),
    lambda: obj.bigreply(),
    lambda: obj.manyargs(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15),
    lambda: obj.noreply([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]),
    lambda: obj.noreply(None),
    lambda: obj.varargs('een', 2, (3,), [4]),
    lambda: obj.keywords(arg1='zork'),
    lambda: obj.echo('een', 2, (3,), [4]),
    lambda: obj.echo({"aap": 42, "noot": 99, "mies": 987654}),
    lambda: obj.bigarg('Argument' * 50),
    lambda: obj.oneway('stringetje', 432423434, 9.8765432)
]

print('-------- BENCHMARK REMOTE OBJECT ---------')
begin = time.time()
iters = 1000
print("warmup...")
for _ in range(iters):
    funcs[0]()
for i, f in enumerate(funcs, start=1):
    print("call #%d, %d times... " % (i, iters), end="")
    before = time.time()
    for _ in range(iters):
        f()
    print("%.3f" % (time.time() - before))
duration = time.time() - begin
print('total time %.3f seconds' % duration)
amount = len(funcs) * iters
print('total method calls: %d' % amount)
avg_pyro_msec = 1000.0 * duration / amount
print('avg. time per method call: %.3f msec (%d/sec) (serializer: %s)' % (avg_pyro_msec, amount / duration, Pyro4.config.SERIALIZER))



from __future__ import print_function
import sys
import time
import warnings

import Pyro4
import serpent


warnings.filterwarnings("ignore")

# Pyro4.config.COMMTIMEOUT=2

print("Enter the server's uri that was printed:")
if sys.version_info < (3, 0):
    uri = raw_input()
else:
    uri = input()
uri = uri.strip()

datasize = 5 * 1024 * 1024  # 5 mb


def do_test(data):
    assert len(data) == datasize
    totalsize = 0

    with Pyro4.core.Proxy(uri) as obj:
        obj._pyroBind()

        begin = time.time()
        for i in range(10):
            print("transferring %d bytes" % datasize)
            size = obj.transfer(data)
            assert size == datasize
            totalsize += datasize
        duration = time.time() - begin

        totalsize = float(totalsize)
        print("It took %.2f seconds to transfer %d mb." % (duration, totalsize / 1024 / 1024))
        print("That is %.0f kb/sec. = %.1f mb/sec. (serializer: %s)" % (totalsize / 1024 / duration, totalsize / 1024 / 1024 / duration, Pyro4.config.SERIALIZER))


def do_test_chunks():
    with Pyro4.core.Proxy(uri) as p:
        totalsize = 0
        begin = time.time()
        for chunk in p.download_chunks(datasize*10):
            chunk = serpent.tobytes(chunk)  # in case of serpent encoded bytes
            totalsize += len(chunk)
            print(".", end="", flush=True)
        assert totalsize == datasize*10
        duration = time.time() - begin
        totalsize = float(totalsize)
        print("It took %.2f seconds to transfer %d mb." % (duration, totalsize / 1024 / 1024))
        print("That is %.0f kb/sec. = %.1f mb/sec. (serializer: %s)" % (totalsize / 1024 / duration, totalsize / 1024 / 1024 / duration, Pyro4.config.SERIALIZER))


data = 'x' * datasize
print("\n\n----test with string data----")
do_test(data)
print("\n\n----test with byte data----")
data = b'x' * datasize
do_test(data)
data = bytearray(b'x' * datasize)
print("\n\n----test with bytearray data----")
do_test(data)
print("\n\n----test download via iterator----")
do_test_chunks()
print("\n\n   (tip: also see the 'filetransfer' example for more efficient ways to transfer large amoungs of binary data)")


from __future__ import print_function
import Pyro4
import Pyro4.constants


class CustomDaemon(Pyro4.Daemon):
    def clientDisconnect(self, conn):
        # If required, you *can* override this to do custom resource freeing.
        # But this is not needed if your resource objects have a proper 'close' method;
        # this method is called by Pyro itself once the client connection gets closed.
        # In this example this override is only used to print out some info.
        print("client disconnects:", conn.sock.getpeername())
        print("    resources: ", [r.name for r in conn.tracked_resources])


class Resource(object):
    # a fictional resource that gets allocated and must be freed again later.
    def __init__(self, name, collection):
        self.name = name
        self.collection = collection

    def close(self):
        # Pyro will call this on a tracked resource once the client's connection gets closed!
        # (Unless the resource can be carbage collected normally by Python.)
        print("Resource: closing", self.name)
        self.collection.discard(self)


@Pyro4.expose
@Pyro4.behavior(instance_mode="single")
class Service(object):
    def __init__(self):
        self.resources = set()      # the allocated resources

    def allocate(self, name):
        resource = Resource(name, self.resources)
        self.resources.add(resource)
        Pyro4.current_context.track_resource(resource)
        print("service: allocated resource", name, " for client", Pyro4.current_context.client_sock_addr)

    def free(self, name):
        resources = {r for r in self.resources if r.name == name}
        self.resources -= resources
        for r in resources:
            r.close()
            Pyro4.current_context.untrack_resource(r)

    def list(self):
        return [r.name for r in self.resources]


with CustomDaemon() as daemon:
    Pyro4.Daemon.serveSimple({
        Service: "service"
    }, ns=False, daemon=daemon)



from __future__ import print_function

import Pyro4
import Pyro4.socketutil
import serpent


# Pyro4.config.COMMTIMEOUT=2


class Testclass(object):
    @Pyro4.expose
    def transfer(self, data):
        if Pyro4.config.SERIALIZER == "serpent" and type(data) is dict:
            data = serpent.tobytes(data)  # in case of serpent encoded bytes
        print("received %d bytes" % len(data))
        return len(data)

    @Pyro4.expose
    def download_chunks(self, size):
        print("client requests a 'streaming' download of %d bytes" % size)
        data = bytearray(size)
        i = 0
        chunksize = 200000
        print("  using chunks of size", chunksize)
        while i < size:
            yield data[i:i+chunksize]
            i += chunksize


Pyro4.Daemon.serveSimple(
    {
        Testclass: "example.hugetransfer"
    },
    host=Pyro4.socketutil.getIpAddress("localhost", workaround127=True),
    ns=False, verbose=True)



"""
Tests for the name server (online/running).

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import time
import threading
import unittest
import Pyro4.core
import Pyro4.naming
import Pyro4.socketutil
import Pyro4.constants
from Pyro4.errors import CommunicationError, NamingError
from Pyro4.configuration import config


class NSLoopThread(threading.Thread):
    def __init__(self, nameserver):
        super(NSLoopThread, self).__init__()
        self.setDaemon(True)
        self.nameserver = nameserver
        self.running = threading.Event()
        self.running.clear()

    def run(self):
        self.running.set()
        try:
            self.nameserver.requestLoop()
        except CommunicationError:
            pass  # ignore pyro communication errors


class BCSetupTests(unittest.TestCase):
    def testBCstart(self):
        myIpAddress = Pyro4.socketutil.getIpAddress("", workaround127=True)
        nsUri, nameserver, bcserver = Pyro4.naming.startNS(host=myIpAddress, port=0, bcport=0, enableBroadcast=False)
        self.assertIsNone(bcserver)
        nameserver.close()
        nsUri, nameserver, bcserver = Pyro4.naming.startNS(host=myIpAddress, port=0, bcport=0, enableBroadcast=True)
        self.assertIsNotNone(bcserver, "expected a BC server to be running. Check DNS setup (hostname must not resolve to loopback address")
        self.assertGreater(bcserver.fileno(), 1)
        self.assertIsNotNone(bcserver.sock)
        nameserver.close()
        bcserver.close()


class NameServerTests(unittest.TestCase):
    def setUp(self):
        config.POLLTIMEOUT = 0.1
        myIpAddress = Pyro4.socketutil.getIpAddress("", workaround127=True)
        self.nsUri, self.nameserver, self.bcserver = Pyro4.naming.startNS(host=myIpAddress, port=0, bcport=0)
        self.assertIsNotNone(self.bcserver, "expected a BC server to be running")
        self.bcserver.runInThread()
        self.daemonthread = NSLoopThread(self.nameserver)
        self.daemonthread.start()
        self.daemonthread.running.wait()
        time.sleep(0.05)
        self.old_bcPort = config.NS_BCPORT
        self.old_nsPort = config.NS_PORT
        self.old_nsHost = config.NS_HOST
        config.NS_PORT = self.nsUri.port
        config.NS_HOST = myIpAddress
        config.NS_BCPORT = self.bcserver.getPort()

    def tearDown(self):
        time.sleep(0.01)
        self.nameserver.shutdown()
        self.bcserver.close()
        self.daemonthread.join()
        config.NS_HOST = self.old_nsHost
        config.NS_PORT = self.old_nsPort
        config.NS_BCPORT = self.old_bcPort

    def testLookupUnixsockParsing(self):
        # this must not raise AttributeError, it did before because of a parse bug
        with self.assertRaises(NamingError):
            Pyro4.naming.locateNS("./u:/tmp/pyro4-naming.usock")

    def testLookupAndRegister(self):
        ns = Pyro4.naming.locateNS()  # broadcast lookup
        self.assertIsInstance(ns, Pyro4.core.Proxy)
        ns._pyroRelease()
        ns = Pyro4.naming.locateNS(self.nsUri.host)  # normal lookup
        self.assertIsInstance(ns, Pyro4.core.Proxy)
        uri = ns._pyroUri
        self.assertEqual("PYRO", uri.protocol)
        self.assertEqual(self.nsUri.host, uri.host)
        self.assertEqual(config.NS_PORT, uri.port)
        self.assertIsNone(ns._pyroHmacKey)
        ns._pyroRelease()
        ns = Pyro4.naming.locateNS(self.nsUri.host, config.NS_PORT, hmac_key=None)
        uri = ns._pyroUri
        self.assertEqual("PYRO", uri.protocol)
        self.assertEqual(self.nsUri.host, uri.host)
        self.assertEqual(config.NS_PORT, uri.port)
        self.assertIsNone(ns._pyroHmacKey)
        # check that we cannot register a stupid type
        self.assertRaises(TypeError, ns.register, "unittest.object1", 5555)
        # we can register str or URI, lookup always returns URI
        ns.register("unittest.object2", "PYRO:55555@host.com:4444")
        self.assertEqual(Pyro4.core.URI("PYRO:55555@host.com:4444"), ns.lookup("unittest.object2"))
        ns.register("unittest.object3", Pyro4.core.URI("PYRO:66666@host.com:4444"))
        self.assertEqual(Pyro4.core.URI("PYRO:66666@host.com:4444"), ns.lookup("unittest.object3"))
        ns._pyroRelease()

    def testLookupInvalidHmac(self):
        with self.assertRaises(NamingError):
            Pyro4.naming.locateNS(self.nsUri.host, config.NS_PORT, hmac_key="invalidkey")

    def testDaemonPyroObj(self):
        uri = self.nsUri
        uri.object = Pyro4.constants.DAEMON_NAME
        with Pyro4.core.Proxy(uri) as daemonobj:
            daemonobj.ping()
            daemonobj.registered()
            try:
                daemonobj.shutdown()
                self.fail("should not succeed to call unexposed method on daemon")
            except AttributeError:
                pass

    def testMulti(self):
        uristr = str(self.nsUri)
        p = Pyro4.core.Proxy(uristr)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro4.naming.resolve(uristr)
        p = Pyro4.core.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro4.naming.resolve(uristr)
        p = Pyro4.core.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro4.naming.resolve(uristr)
        p = Pyro4.core.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro4.naming.resolve(uristr)
        p = Pyro4.core.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro4.naming.resolve(uristr)
        p = Pyro4.core.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        daemonUri = "PYRO:" + Pyro4.constants.DAEMON_NAME + "@" + uri.location
        _ = Pyro4.naming.resolve(daemonUri)
        _ = Pyro4.naming.resolve(daemonUri)
        _ = Pyro4.naming.resolve(daemonUri)
        _ = Pyro4.naming.resolve(daemonUri)
        _ = Pyro4.naming.resolve(daemonUri)
        _ = Pyro4.naming.resolve(daemonUri)
        uri = Pyro4.naming.resolve(daemonUri)
        pyronameUri = "PYRONAME:" + Pyro4.constants.NAMESERVER_NAME + "@" + uri.location
        _ = Pyro4.naming.resolve(pyronameUri)
        _ = Pyro4.naming.resolve(pyronameUri)
        _ = Pyro4.naming.resolve(pyronameUri)
        _ = Pyro4.naming.resolve(pyronameUri)
        _ = Pyro4.naming.resolve(pyronameUri)
        _ = Pyro4.naming.resolve(pyronameUri)

    def testResolve(self):
        resolved1 = Pyro4.naming.resolve(Pyro4.core.URI("PYRO:12345@host.com:4444"), hmac_key=None)
        resolved2 = Pyro4.naming.resolve("PYRO:12345@host.com:4444", hmac_key=None)
        self.assertTrue(type(resolved1) is Pyro4.core.URI)
        self.assertEqual(resolved1, resolved2)
        self.assertEqual("PYRO:12345@host.com:4444", str(resolved1))

        ns = Pyro4.naming.locateNS(self.nsUri.host, self.nsUri.port)
        host = "[" + self.nsUri.host + "]" if ":" in self.nsUri.host else self.nsUri.host
        uri = Pyro4.naming.resolve("PYRONAME:" + Pyro4.constants.NAMESERVER_NAME + "@" + host + ":" + str(self.nsUri.port))
        self.assertEqual("PYRO", uri.protocol)
        self.assertEqual(self.nsUri.host, uri.host)
        self.assertEqual(Pyro4.constants.NAMESERVER_NAME, uri.object)
        self.assertEqual(uri, ns._pyroUri)
        ns._pyroRelease()

        # broadcast lookup
        self.assertRaises(NamingError, Pyro4.naming.resolve, "PYRONAME:unknown_object")
        uri = Pyro4.naming.resolve("PYRONAME:" + Pyro4.constants.NAMESERVER_NAME)
        self.assertEqual(Pyro4.core.URI, type(uri))
        self.assertEqual("PYRO", uri.protocol)

        # test some errors
        self.assertRaises(NamingError, Pyro4.naming.resolve, "PYRONAME:unknown_object@" + host)
        self.assertRaises(TypeError, Pyro4.naming.resolve, 999)  # wrong arg type

    def testRefuseDottedNames(self):
        old_metadata = config.METADATA
        config.METADATA = False
        with Pyro4.naming.locateNS(self.nsUri.host, self.nsUri.port) as ns:
            # the name server should never have dotted names enabled
            self.assertRaises(AttributeError, ns.namespace.keys)
            self.assertIsNotNone(ns._pyroConnection)
        self.assertIsNone(ns._pyroConnection)
        config.METADATA = old_metadata

    def testAutoClean(self):
        try:
            config.NS_AUTOCLEAN = 0.0
            config.COMMTIMEOUT = 0.5
            Pyro4.naming.AutoCleaner.max_unreachable_time = 1
            Pyro4.naming.AutoCleaner.loop_delay = 0.5
            Pyro4.naming.AutoCleaner.override_autoclean_min = True
            with Pyro4.naming.NameServerDaemon(port=0) as ns:
                self.assertIsNone(ns.cleaner_thread)
            config.NS_AUTOCLEAN = 0.2
            with Pyro4.naming.NameServerDaemon(port=0) as ns:
                self.assertIsNotNone(ns.cleaner_thread)
                ns.nameserver.register("test", "PYRO:test@localhost:59999")
                self.assertEqual(2, ns.nameserver.count())
                time.sleep(4)
                self.assertEqual(1, ns.nameserver.count(), "registration should be cleaned up")
            self.assertIsNone(ns.cleaner_thread)
        finally:
            Pyro4.naming.AutoCleaner.override_autoclean_min = False
            Pyro4.naming.AutoCleaner.max_unreachable_time = 20
            Pyro4.naming.AutoCleaner.loop_delay = 2
            config.NS_AUTOCLEAN = 0.0
            config.COMMTIMEOUT = 0.0


class NameServerTests0000(unittest.TestCase):
    def setUp(self):
        config.POLLTIMEOUT = 0.1
        self.nsUri, self.nameserver, self.bcserver = Pyro4.naming.startNS(host="", port=0, bcport=0)
        host_check = self.nsUri.host
        self.assertEqual("0.0.0.0", host_check, "for hostname \"\" the resulting ip must be 0.0.0.0 (or ipv6 equivalent)")
        self.assertIsNotNone(self.bcserver, "expected a BC server to be running")
        self.bcthread = self.bcserver.runInThread()
        self.old_bcPort = config.NS_BCPORT
        self.old_nsPort = config.NS_PORT
        self.old_nsHost = config.NS_HOST
        config.NS_PORT = self.nsUri.port
        config.NS_HOST = self.nsUri.host
        config.NS_BCPORT = self.bcserver.getPort()

    def tearDown(self):
        time.sleep(0.01)
        self.nameserver.shutdown()
        self.bcserver.close()
        self.bcthread.join()
        config.NS_HOST = self.old_nsHost
        config.NS_PORT = self.old_nsPort
        config.NS_BCPORT = self.old_bcPort

    def testBCLookup0000(self):
        ns = Pyro4.naming.locateNS()  # broadcast lookup
        self.assertIsInstance(ns, Pyro4.core.Proxy)
        self.assertNotEqual("0.0.0.0", ns._pyroUri.host, "returned location must not be 0.0.0.0 when running on 0.0.0.0")
        ns._pyroRelease()


class NameServerTestsHmac(unittest.TestCase):
    def setUp(self):
        config.POLLTIMEOUT = 0.1
        myIpAddress = Pyro4.socketutil.getIpAddress("", workaround127=True)
        self.nsUri, self.nameserver, self.bcserver = Pyro4.naming.startNS(host=myIpAddress, port=0, bcport=0, hmac=b"test_key")
        self.assertIsNotNone(self.bcserver, "expected a BC server to be running")
        self.bcserver.runInThread()
        self.daemonthread = NSLoopThread(self.nameserver)
        self.daemonthread.start()
        self.daemonthread.running.wait()
        time.sleep(0.05)
        self.old_bcPort = config.NS_BCPORT
        self.old_nsPort = config.NS_PORT
        self.old_nsHost = config.NS_HOST
        config.NS_PORT = self.nsUri.port
        config.NS_HOST = myIpAddress
        config.NS_BCPORT = self.bcserver.getPort()

    def tearDown(self):
        time.sleep(0.01)
        self.nameserver.shutdown()
        self.bcserver.close()
        self.daemonthread.join()
        config.NS_HOST = self.old_nsHost
        config.NS_PORT = self.old_nsPort
        config.NS_BCPORT = self.old_bcPort

    def testLookupAndRegister(self):
        ns = Pyro4.naming.locateNS()  # broadcast lookup without providing hmac still works
        self.assertIsInstance(ns, Pyro4.core.Proxy)
        self.assertIsNone(ns._pyroHmacKey)   #... but no hmac is set on the proxy
        ns._pyroRelease()
        ns = Pyro4.naming.locateNS(hmac_key=b"test_key")  # broadcast lookup providing hmac
        self.assertIsInstance(ns, Pyro4.core.Proxy)
        self.assertEqual(b"test_key", ns._pyroHmacKey)  # ... sets the hmac on the proxy
        ns._pyroRelease()
        ns = Pyro4.naming.locateNS(self.nsUri.host, config.NS_PORT, hmac_key=b"test_key")
        uri = ns._pyroUri
        self.assertEqual("PYRO", uri.protocol)
        self.assertEqual(self.nsUri.host, uri.host)
        self.assertEqual(config.NS_PORT, uri.port)
        self.assertEqual(b"test_key", ns._pyroHmacKey)
        ns._pyroRelease()

    def testResolve(self):
        uri = Pyro4.naming.resolve("PYRONAME:Pyro.NameServer", hmac_key=b"test_key")
        self.assertEqual("PYRO", uri.protocol)
        self.assertEqual(self.nsUri.host, uri.host)
        self.assertEqual(config.NS_PORT, uri.port)

    def testPyroname(self):
        with Pyro4.core.Proxy("PYRONAME:Pyro.NameServer") as p:
            p._pyroHmacKey = b"test_key"
            p.ping()   # the resolve() that is done should also use the hmac key

    def testResolveWrongHmac(self):
        with self.assertRaises(CommunicationError) as ex:
            Pyro4.naming.resolve("PYRONAME:Pyro.NameServer", hmac_key=b"wrong_key")
        emsg = str(ex.exception)
        self.assertTrue(emsg.startswith("cannot connect to "))
        self.assertTrue(emsg.endswith("message hmac mismatch"))

    def testResolveAsymmetricHmacUsage(self):
        with self.assertRaises(CommunicationError) as ex:
            Pyro4.naming.resolve("PYRONAME:Pyro.NameServer", hmac_key=None)
        emsg = str(ex.exception)
        self.assertTrue(emsg.startswith("cannot connect to "))
        self.assertTrue(emsg.endswith("hmac key config not symmetric"))


if __name__ == "__main__":
    unittest.main()



from __future__ import print_function
import sys
import Pyro4
from listener import Listener


if len(sys.argv) != 2:
    print("Give topic as argument.")
else:
    topic = sys.argv[1].strip()
    if not topic:
        raise ValueError("Must give topic name.")
    listener = Listener(topic)
    daemon = Pyro4.Daemon()
    daemon.register(listener)
    listener.register_with_dispatcher()
    print("Listener for topic {} waiting for data.".format(topic))
    daemon.requestLoop()



from __future__ import print_function

try:
    import queue
except ImportError:
    import Queue as queue
import Pyro4
from Pyro4.util import SerializerBase
from workitem import Workitem


# For 'workitem.Workitem' we register a deserialization hook to be able to get these back from Pyro
SerializerBase.register_dict_to_class("workitem.Workitem", Workitem.from_dict)


@Pyro4.expose
@Pyro4.behavior(instance_mode="single")
class DispatcherQueue(object):
    def __init__(self):
        self.workqueue = queue.Queue()
        self.resultqueue = queue.Queue()

    def putWork(self, item):
        self.workqueue.put(item)

    def getWork(self, timeout=5):
        try:
            return self.workqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            raise ValueError("no items in queue")

    def putResult(self, item):
        self.resultqueue.put(item)

    def getResult(self, timeout=5):
        try:
            return self.resultqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            raise ValueError("no result available")

    def workQueueSize(self):
        return self.workqueue.qsize()

    def resultQueueSize(self):
        return self.resultqueue.qsize()

# main program

Pyro4.Daemon.serveSimple({
    DispatcherQueue: "example.distributed.dispatcher"
})



from __future__ import print_function
import Pyro4


def myfunction(a, b, extra=None):
    print(">>> myfunction called with: a={0}, b={1}, extra={2}".format(a, b, extra))
    return a + b


print("\n* just a single future call:")
future = Pyro4.Future(myfunction)
result = future(5, 6)
# we can do stuff here in the meantime...
print("result value=", result.value)
assert result.value == 11

print("\n* several calls chained:")
future = Pyro4.Future(myfunction)
future.then(myfunction, 10)
future.then(myfunction, 20, extra="something")
# the callables will be invoked like so:  function(asyncvalue, normalarg, kwarg=..., kwarg=...)
# (the value from the previous call is passed as the first argument to the next call)
result = future(5, 6)
# we can do stuff here in the meantime...
print("result value=", result.value)
assert result.value == 41



from __future__ import print_function
import Pyro4

something = "Something"


@Pyro4.expose
class Thingy(object):
    def __init__(self):
        self.sub = {"name": "value"}
        self.value = 42
        self._value = 123
        self.__value = 999

    def __dunder__(self):
        return "yep"

    def __len__(self):
        return 200

    def getValue(self):
        return self.value

    @property
    def prop_value(self):
        return self.value

    @prop_value.setter
    def prop_value(self, value):
        self.value = value

    @property
    def prop_sub(self):
        return self.sub


Pyro4.Daemon.serveSimple({
    Thingy: "example.attributes"
}, ns=False)



from __future__ import print_function
import time

import Pyro4


print("Autoreconnect using Name Server.")

# We create a proxy with a PYRONAME uri.
# That allows Pyro to look up the object again in the NS when
# it needs to reconnect later.
obj = Pyro4.core.Proxy("PYRONAME:example.autoreconnect")

while True:
    print("call...")
    try:
        obj.method(42)
        print("Sleeping 1 second")
        time.sleep(1)
    except Pyro4.errors.ConnectionClosedError:  # or possibly CommunicationError
        print("Connection lost. REBINDING...")
        print("(restart the server now)")
        obj._pyroReconnect()



"""
This is the publisher meant for the 'weather' messages example.
"""
from __future__ import print_function
import time
import random
import sys
import Pyro4
import Pyro4.errors
from Pyro4.util import excepthook
from messagebus import PYRO_MSGBUS_NAME

sys.excepthook = excepthook
if sys.version_info < (3, 0):
    input = raw_input


location = input("Give city or country to use as location: ").strip() or 'Amsterdam'

bus = Pyro4.Proxy("PYRONAME:"+PYRO_MSGBUS_NAME)
bus.add_topic("weather-forecast")

while True:
    time.sleep(0.01)
    forecast = (location, random.choice(["sunny", "cloudy", "storm", "rainy", "hail", "thunder", "calm", "mist", "cold", "hot"]))
    print("Forecast:", forecast)
    try:
        bus.send("weather-forecast", forecast)
    except Pyro4.errors.CommunicationError:
        print("connection to the messagebus is lost, reconnecting...")
        bus._pyroReconnect()



"""
Tests for the utility functions.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import sys
import os
import unittest
import Pyro4.util
from Pyro4.configuration import config
from testsupport import *


# noinspection PyUnusedLocal
def crash(arg=100):
    pre1 = "black"
    pre2 = 999

    # noinspection PyUnusedLocal
    def nest(p1, p2):
        q = "white" + pre1
        x = pre2
        y = arg // 2
        p3 = p1 // p2
        return p3

    a = 10
    b = 0
    s = "hello"
    c = nest(a, b)
    return c


class TestUtils(unittest.TestCase):
    def testFormatTracebackNormal(self):
        try:
            crash()
            self.fail("must crash with ZeroDivisionError")
        except ZeroDivisionError:
            tb = "".join(Pyro4.util.formatTraceback(detailed=False))
            self.assertIn("p3 = p1 // p2", tb)
            self.assertIn("ZeroDivisionError", tb)
            self.assertNotIn(" a = 10", tb)
            self.assertNotIn(" s = 'whiteblack'", tb)
            self.assertNotIn(" pre2 = 999", tb)
            self.assertNotIn(" x = 999", tb)

    def testFormatTracebackDetail(self):
        try:
            crash()
            self.fail("must crash with ZeroDivisionError")
        except ZeroDivisionError:
            tb = "".join(Pyro4.util.formatTraceback(detailed=True))
            self.assertIn("p3 = p1 // p2", tb)
            self.assertIn("ZeroDivisionError", tb)
            if sys.platform != "cli":
                self.assertIn(" a = 10", tb)
                self.assertIn(" q = 'whiteblack'", tb)
                self.assertIn(" pre2 = 999", tb)
                self.assertIn(" x = 999", tb)

    def testPyroTraceback(self):
        try:
            crash()
            self.fail("must crash with ZeroDivisionError")
        except ZeroDivisionError:
            pyro_tb = Pyro4.util.formatTraceback(detailed=True)
            if sys.platform != "cli":
                self.assertIn(" Extended stacktrace follows (most recent call last)\n", pyro_tb)
        try:
            crash("stringvalue")
            self.fail("must crash with TypeError")
        except TypeError as x:
            x._pyroTraceback = pyro_tb  # set the remote traceback info
            pyrotb = "".join(Pyro4.util.getPyroTraceback())
            self.assertIn("Remote traceback", pyrotb)
            self.assertIn("crash(\"stringvalue\")", pyrotb)
            self.assertIn("TypeError:", pyrotb)
            self.assertIn("ZeroDivisionError", pyrotb)
            del x._pyroTraceback
            pyrotb = "".join(Pyro4.util.getPyroTraceback())
            self.assertNotIn("Remote traceback", pyrotb)
            self.assertNotIn("ZeroDivisionError", pyrotb)
            self.assertIn("crash(\"stringvalue\")", pyrotb)
            self.assertIn("TypeError:", pyrotb)

    def testPyroTracebackArgs(self):
        try:
            crash()
            self.fail("must crash with ZeroDivisionError")
        except ZeroDivisionError:
            ex_type, ex_value, ex_tb = sys.exc_info()
            x = ex_value
            tb1 = Pyro4.util.getPyroTraceback()
            tb2 = Pyro4.util.getPyroTraceback(ex_type, ex_value, ex_tb)
            self.assertEqual(tb1, tb2)
            tb1 = Pyro4.util.formatTraceback()
            tb2 = Pyro4.util.formatTraceback(ex_type, ex_value, ex_tb)
            self.assertEqual(tb1, tb2)
            tb2 = Pyro4.util.formatTraceback(detailed=True)
            if sys.platform != "cli":
                self.assertNotEqual(tb1, tb2)
            # old call syntax, should get an error now:
            self.assertRaises(TypeError, Pyro4.util.getPyroTraceback, x)
            self.assertRaises(TypeError, Pyro4.util.formatTraceback, x)

    def testExcepthook(self):
        # simply test the excepthook by calling it the way Python would
        try:
            crash()
            self.fail("must crash with ZeroDivisionError")
        except ZeroDivisionError:
            pyro_tb = Pyro4.util.formatTraceback()
        try:
            crash("stringvalue")
            self.fail("must crash with TypeError")
        except TypeError:
            ex_type, ex_value, ex_tb = sys.exc_info()
            ex_value._pyroTraceback = pyro_tb  # set the remote traceback info
            oldstderr = sys.stderr
            try:
                sys.stderr = StringIO()
                Pyro4.util.excepthook(ex_type, ex_value, ex_tb)
                output = sys.stderr.getvalue()
                self.assertIn("Remote traceback", output)
                self.assertIn("crash(\"stringvalue\")", output)
                self.assertIn("TypeError:", output)
                self.assertIn("ZeroDivisionError", output)
            finally:
                sys.stderr = oldstderr

    def clearEnv(self):
        if "PYRO_HOST" in os.environ:
            del os.environ["PYRO_HOST"]
        if "PYRO_NS_PORT" in os.environ:
            del os.environ["PYRO_NS_PORT"]
        if "PYRO_COMPRESSION" in os.environ:
            del os.environ["PYRO_COMPRESSION"]
        config.reset()

    def testConfig(self):
        self.clearEnv()
        try:
            self.assertEqual(9090, config.NS_PORT)
            self.assertEqual("localhost", config.HOST)
            self.assertEqual(False, config.COMPRESSION)
            os.environ["NS_PORT"] = "4444"
            config.reset()
            self.assertEqual(9090, config.NS_PORT)
            os.environ["PYRO_NS_PORT"] = "4444"
            os.environ["PYRO_HOST"] = "something.com"
            os.environ["PYRO_COMPRESSION"] = "OFF"
            config.reset()
            self.assertEqual(4444, config.NS_PORT)
            self.assertEqual("something.com", config.HOST)
            self.assertEqual(False, config.COMPRESSION)
        finally:
            self.clearEnv()
            self.assertEqual(9090, config.NS_PORT)
            self.assertEqual("localhost", config.HOST)
            self.assertEqual(False, config.COMPRESSION)

    def testConfigReset(self):
        try:
            config.reset()
            self.assertEqual("localhost", config.HOST)
            config.HOST = "foobar"
            self.assertEqual("foobar", config.HOST)
            config.reset()
            self.assertEqual("localhost", config.HOST)
            os.environ["PYRO_HOST"] = "foobar"
            config.reset()
            self.assertEqual("foobar", config.HOST)
            del os.environ["PYRO_HOST"]
            config.reset()
            self.assertEqual("localhost", config.HOST)
        finally:
            self.clearEnv()

    def testResolveAttr(self):
        config.REQUIRE_EXPOSE = True
        @Pyro4.core.expose
        class Exposed(object):
            def __init__(self, value):
                self.propvalue = value
                self.__value__ = value   # is not affected by the @expose

            def __str__(self):
                return "<%s>" % self.value

            def _p(self):
                return "should not be allowed"

            def __p(self):
                return "should not be allowed"

            def __p__(self):
                return "should be allowed (dunder)"

            @property
            def value(self):
                return self.propvalue

        class Unexposed(object):
            def __init__(self):
                self.value = 42

            def __value__(self):
                return self.value

        obj = Exposed("hello")
        obj.a = Exposed("a")
        obj.a.b = Exposed("b")
        obj.a.b.c = Exposed("c")
        obj.a._p = Exposed("p1")
        obj.a._p.q = Exposed("q1")
        obj.a.__p = Exposed("p2")
        obj.a.__p.q = Exposed("q2")
        obj.u = Unexposed()
        obj.u.v = Unexposed()
        # check the accessible attributes
        self.assertEqual("<a>", str(Pyro4.util.getAttribute(obj, "a")))
        dunder = str(Pyro4.util.getAttribute(obj, "__p__"))
        self.assertTrue(dunder.startswith("<bound method "))  # dunder is not private, part 1 of the check
        self.assertTrue("Exposed.__p__ of" in dunder)  # dunder is not private, part 2 of the check
        # check what should not be accessible
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "value")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "propvalue")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "__value__")  # is not affected by the @expose
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "_p")  # private
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "__p")  # private
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "a.b")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "a.b.c")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "a.b.c.d")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "a._p")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "a._p.q")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "a.__p.q")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "u")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "u.v")
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "u.v.value")
        # try again but with require expose disabled
        config.REQUIRE_EXPOSE = False
        self.assertIsInstance(Pyro4.util.getAttribute(obj, "u"), Unexposed)
        self.assertEqual("<a>", str(Pyro4.util.getAttribute(obj, "a")))
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "u.v")  # still not allowed to follow the dots
        self.assertRaises(AttributeError, Pyro4.util.getAttribute, obj, "u.v.value")  # still not allowed to follow the dots
        config.REQUIRE_EXPOSE = True

    def testUnicodeKwargs(self):
        # test the way the interpreter deals with unicode function kwargs
        def function(*args, **kwargs):
            return args, kwargs

        processed_args = function(*(1, 2, 3), **{unichr(65): 42})
        self.assertEqual(((1, 2, 3), {unichr(65): 42}), processed_args)
        processed_args = function(*(1, 2, 3), **{unichr(0x20ac): 42})
        key = list(processed_args[1].keys())[0]
        self.assertEqual(key, unichr(0x20ac))
        self.assertEqual(((1, 2, 3), {unichr(0x20ac): 42}), processed_args)


class TestMetaAndExpose(unittest.TestCase):
    def setUp(self):
        config.REQUIRE_EXPOSE = True

    def testBasic(self):
        o = MyThingFullExposed("irmen")
        m1 = Pyro4.util.get_exposed_members(o)
        m2 = Pyro4.util.get_exposed_members(MyThingFullExposed)
        self.assertEqual(m1, m2)
        keys = m1.keys()
        self.assertEqual(3, len(keys))
        self.assertIn("methods", keys)
        self.assertIn("attrs", keys)
        self.assertIn("oneway", keys)

    def testResetExposedCache(self):
        o = MyThingFullExposed("irmen")
        m1 = Pyro4.util.get_exposed_members(o)
        self.assertNotIn("newly_added_method", m1["methods"])
        MyThingFullExposed.newly_added_method = Pyro4.core.expose(lambda self: None)
        m2 = Pyro4.util.get_exposed_members(o)
        self.assertNotIn("newly_added_method", m2["methods"])
        Pyro4.util.reset_exposed_members(o)
        m3 = Pyro4.util.get_exposed_members(o)
        self.assertIn("newly_added_method", m3["methods"])

    def testGetExposedCacheWorks(self):
        class Thingy(object):
            def method1(self):
                pass
            @property
            def prop(self):
                return 1
            def notexposed(self):
                pass
        m1 = Pyro4.util.get_exposed_members(Thingy, only_exposed=False)
        def new_method(self, arg):
            return arg
        Thingy.new_method = new_method
        m2 = Pyro4.util.get_exposed_members(Thingy, only_exposed=False)
        self.assertEqual(m1, m2, "should still be equal because result from cache")
        m2 = Pyro4.util.get_exposed_members(Thingy, only_exposed=False, use_cache=False)
        self.assertNotEqual(m1, m2, "should not be equal because new result not from cache")

    def testPrivateNotExposed(self):
        o = MyThingFullExposed("irmen")
        m = Pyro4.util.get_exposed_members(o)
        self.assertEqual({"classmethod", "staticmethod", "method", "__dunder__", "oneway", "exposed"}, m["methods"])
        self.assertEqual({"prop1", "readonly_prop1", "prop2"}, m["attrs"])
        self.assertEqual({"oneway"}, m["oneway"])
        o = MyThingPartlyExposed("irmen")
        m = Pyro4.util.get_exposed_members(o)
        self.assertEqual({"oneway", "exposed"}, m["methods"])
        self.assertEqual({"prop1", "readonly_prop1"}, m["attrs"])
        self.assertEqual({"oneway"}, m["oneway"])

    def testNotOnlyExposed(self):
        o = MyThingPartlyExposed("irmen")
        m = Pyro4.util.get_exposed_members(o, only_exposed=False)
        self.assertEqual({"classmethod", "staticmethod", "method", "__dunder__", "oneway", "exposed"}, m["methods"])
        self.assertEqual({"prop1", "readonly_prop1", "prop2"}, m["attrs"])
        self.assertEqual({"oneway"}, m["oneway"])

    def testPartlyExposedSubclass(self):
        o = MyThingPartlyExposedSub("irmen")
        m = Pyro4.util.get_exposed_members(o)
        self.assertEqual({"prop1", "readonly_prop1"}, m["attrs"])
        self.assertEqual({"oneway"}, m["oneway"])
        self.assertEqual({"sub_exposed", "exposed", "oneway"}, m["methods"])

    def testExposedSubclass(self):
        o = MyThingExposedSub("irmen")
        m = Pyro4.util.get_exposed_members(o)
        self.assertEqual({"readonly_prop1", "prop1", "prop2"}, m["attrs"])
        self.assertEqual({"oneway", "oneway2"}, m["oneway"])
        self.assertEqual({"classmethod", "staticmethod", "oneway", "__dunder__", "method", "exposed",
                          "oneway2", "sub_exposed", "sub_unexposed"}, m["methods"])

    def testExposePrivateFails(self):
        with self.assertRaises(AttributeError):
            class Test1(object):
                @Pyro4.core.expose
                def _private(self):
                    pass
        with self.assertRaises(AttributeError):
            class Test3(object):
                @Pyro4.core.expose
                def __private(self):
                    pass
        with self.assertRaises(AttributeError):
            @Pyro4.core.expose
            class _Test4(object):
                pass
        with self.assertRaises(AttributeError):
            @Pyro4.core.expose
            class __Test5(object):
                pass

    def testExposeDunderOk(self):
        class Test1(object):
            @Pyro4.core.expose
            def __dunder__(self):
                pass
        self.assertTrue(Test1.__dunder__._pyroExposed)
        @Pyro4.core.expose
        class Test2(object):
            def __dunder__(self):
                pass
        self.assertTrue(Test2._pyroExposed)
        self.assertTrue(Test2.__dunder__._pyroExposed)

    def testClassmethodExposeWrongOrderFail(self):
        with self.assertRaises(AttributeError) as ax:
            class TestClass:
                @Pyro4.core.expose
                @classmethod
                def cmethod(cls):
                    pass
        self.assertTrue("must be done after" in str(ax.exception))
        with self.assertRaises(AttributeError) as ax:
            class TestClass:
                @Pyro4.core.expose
                @staticmethod
                def smethod(cls):
                    pass
        self.assertTrue("must be done after" in str(ax.exception))

    def testClassmethodExposeCorrectOrderOkay(self):
        class TestClass:
            @classmethod
            @Pyro4.core.expose
            def cmethod(cls):
                pass
            @staticmethod
            @Pyro4.core.expose
            def smethod(cls):
                pass
        self.assertTrue(TestClass.cmethod._pyroExposed)
        self.assertTrue(TestClass.smethod._pyroExposed)

    def testGetExposedProperty(self):
        o = MyThingFullExposed("irmen")
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "name")
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "c_attr")
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "propvalue")
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "unexisting_attribute")
        self.assertEqual(42, Pyro4.util.get_exposed_property_value(o, "prop1"))
        self.assertEqual(42, Pyro4.util.get_exposed_property_value(o, "prop2"))

    def testGetExposedPropertyFromPartiallyExposed(self):
        o = MyThingPartlyExposed("irmen")
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "name")
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "c_attr")
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "propvalue")
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "unexisting_attribute")
        self.assertEqual(42, Pyro4.util.get_exposed_property_value(o, "prop1"))
        with self.assertRaises(AttributeError):
            Pyro4.util.get_exposed_property_value(o, "prop2")

    def testSetExposedProperty(self):
        o = MyThingFullExposed("irmen")
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "name", "erorr")
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "unexisting_attribute", 42)
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "readonly_prop1", 42)
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "propvalue", 999)
        self.assertEqual(42, o.prop1)
        self.assertEqual(42, o.prop2)
        Pyro4.util.set_exposed_property_value(o, "prop1", 999)
        self.assertEqual(999, o.propvalue)
        Pyro4.util.set_exposed_property_value(o, "prop2", 8888)
        self.assertEqual(8888, o.propvalue)

    def testSetExposedPropertyFromPartiallyExposed(self):
        o = MyThingPartlyExposed("irmen")
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "name", "erorr")
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "unexisting_attribute", 42)
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "readonly_prop1", 42)
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "propvalue", 999)
        self.assertEqual(42, o.prop1)
        self.assertEqual(42, o.prop2)
        Pyro4.util.set_exposed_property_value(o, "prop1", 999)
        self.assertEqual(999, o.propvalue)
        with self.assertRaises(AttributeError):
            Pyro4.util.set_exposed_property_value(o, "prop2", 8888)

    def testIsPrivateName(self):
        self.assertTrue(Pyro4.util.is_private_attribute("_"))
        self.assertTrue(Pyro4.util.is_private_attribute("__"))
        self.assertTrue(Pyro4.util.is_private_attribute("___"))
        self.assertTrue(Pyro4.util.is_private_attribute("_p"))
        self.assertTrue(Pyro4.util.is_private_attribute("_pp"))
        self.assertTrue(Pyro4.util.is_private_attribute("_p_"))
        self.assertTrue(Pyro4.util.is_private_attribute("_p__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__p"))
        self.assertTrue(Pyro4.util.is_private_attribute("___p"))
        self.assertFalse(Pyro4.util.is_private_attribute("__dunder__"))  # dunder methods should not be private except a list of exceptions as tested below
        self.assertTrue(Pyro4.util.is_private_attribute("__init__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__call__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__new__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__del__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__repr__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__unicode__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__str__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__format__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__nonzero__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__bool__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__coerce__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__cmp__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__eq__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__ne__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__lt__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__gt__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__le__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__ge__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__hash__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__dir__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__enter__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__exit__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__copy__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__deepcopy__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__sizeof__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__getattr__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__setattr__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__hasattr__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__delattr__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__getattribute__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__instancecheck__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__subclasscheck__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__subclasshook__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__getinitargs__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__getnewargs__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__getstate__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__setstate__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__reduce__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__reduce_ex__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__getstate_for_dict__"))
        self.assertTrue(Pyro4.util.is_private_attribute("__setstate_from_dict__"))


if __name__ == "__main__":
    # import sys; sys.argv = ['', 'Test.testName']
    unittest.main()



from __future__ import print_function
import sys

import Pyro4
import Pyro4.util


print("First start the built-in test echo server with something like:")
print("$ python -m Pyro4.test.echoserver")
print("Enter the server's uri that was printed:")
if sys.version_info < (3, 0):
    uri = raw_input()
else:
    uri = input()
uri = uri.strip()
echoserver = Pyro4.Proxy(uri)

response = echoserver.echo("hello")
print("\ngot back from the server: %s" % response)
response = echoserver.echo([1, 2, 3, 4])
print("got back from the server: %s" % response)

for element in echoserver.generator():
    print("got element from remote iterator:", element)

try:
    echoserver.error()
except:
    print("\ncaught an exception (expected), traceback:")
    print("".join(Pyro4.util.getPyroTraceback()))

print("\nshutting down the test echo server. (restart it if you want to run this again)")
echoserver.shutdown()



from __future__ import print_function
import sys
import datetime
import Pyro4.util
import Pyro4.core
import Pyro4
from customdata import CustomData

if sys.version_info < (3, 0):
    input = raw_input


sys.excepthook = Pyro4.util.excepthook

# teach Serpent how to serialize our data class
Pyro4.util.SerializerBase.register_class_to_dict(CustomData, CustomData.to_dict)


with Pyro4.Proxy("PYRONAME:example.blobdispatcher") as dispatcher:
    while True:
        topic = input("Enter topic to send data on (just enter to quit) ").strip()
        if not topic:
            break
        # create our custom data object and send it through the dispatcher
        data = CustomData(42, "hello world", datetime.datetime.now())
        dispatcher.process_blob(Pyro4.core.SerializedBlob(topic, data))
        print("processed")



from __future__ import print_function
import Pyro4
import datetime
import sys

if sys.version_info < (3, 0):
    input = raw_input


uri = input("What is the server uri? ").strip()
fmt = '%Y-%m-%d %H:%M:%S %Z%z'
print("local time without timezone: ", datetime.datetime.now().strftime(fmt))


def test():
    with Pyro4.core.Proxy(uri) as serv:
        print("\nFIRST: no timezone")
        date1 = serv.echo(datetime.datetime.now())
        print("{0}\n  {1} ({2})".format(date1, repr(date1), type(date1)))
        if hasattr(date1, "tzinfo"):
            print("    tzinfo =", date1.tzinfo)
        else:
            print("    no tzinfo attribute")

        print("\nSECOND: PyTz timezones")
        date1 = serv.pytz()
        assert isinstance(date1.tzinfo, datetime.tzinfo)
        print("{0}\n  {1} ({2})\n    {3}".format(date1, date1.tzinfo, type(date1.tzinfo), date1.strftime(fmt)))
        date2 = serv.echo(date1)
        print("{0}\n  {1} ({2})\n    {3}".format(date2, date2.tzinfo, type(date2.tzinfo), date2.strftime(fmt)))
        assert date1 == date2

        print("\nTHIRD: DateUtil timezones")
        date1 = serv.dateutil()
        assert isinstance(date1.tzinfo, datetime.tzinfo)
        print("{0}\n  {1} ({2})\n    {3}".format(date1, date1.tzinfo, type(date1.tzinfo), date1.strftime(fmt)))
        date2 = serv.echo(date1)
        print("{0}\n  {1} ({2})\n    {3}".format(date2, date2.tzinfo, type(date2.tzinfo), date2.strftime(fmt)))
        assert date1 == date2


# pickle.
print("\n******* pickle *******")
print("******* (expecting no errors) ******")
Pyro4.config.SERIALIZER = "pickle"
try:
    test()
except:
    import traceback
    traceback.print_exc()

# serpent.
print("\n******* serpent *******")
Pyro4.config.SERIALIZER = "serpent"
try:
    test()
except:
    import traceback
    traceback.print_exc()

# json.
print("\n******* json *******")
Pyro4.config.SERIALIZER = "json"
try:
    test()
except:
    import traceback
    traceback.print_exc()

# marshal.
print("\n******* marshal *******")
Pyro4.config.SERIALIZER = "marshal"
try:
    test()
except:
    import traceback
    traceback.print_exc()




from __future__ import print_function
import Pyro4


class Viewer(object):
    def __init__(self):
        self.markets = set()
        self.symbols = set()

    def start(self):
        print("Shown quotes:", self.symbols)
        quote_sources = {
            market.name: market.quotes() for market in self.markets
        }
        while True:
            for market, quote_source in quote_sources.items():
                quote = next(quote_source)  # get a new stock quote from the source
                symbol, value = quote
                if symbol in self.symbols:
                    print("{0}.{1}: {2}".format(market, symbol, value))


def find_stockmarkets():
    # You can hardcode the stockmarket names for nasdaq and newyork, but it
    # is more flexible if we just look for every available stockmarket.
    markets = []
    with Pyro4.locateNS() as ns:
        for market, market_uri in ns.list(prefix="example.stockmarket.").items():
            print("found market", market)
            markets.append(Pyro4.Proxy(market_uri))
    if not markets:
        raise ValueError("no markets found! (have you started the stock markets first?)")
    return markets


def main():
    viewer = Viewer()
    viewer.markets = find_stockmarkets()
    viewer.symbols = {"IBM", "AAPL", "MSFT"}
    viewer.start()


if __name__ == "__main__":
    main()



"""
This is the producer for the 'many topics' messages example.
"""
from __future__ import print_function
import time
import random
import sys
import Pyro4
from Pyro4.util import excepthook
from messagebus import PYRO_MSGBUS_NAME

sys.excepthook = excepthook

# add a bunch of topics to the bus
bus = Pyro4.Proxy("PYRONAME:"+PYRO_MSGBUS_NAME)
for letter in "abcdefghijklmnopqrstuvwxyz":
    bus.add_topic("msg.topic.%s" % letter)

print("publishing messages on a random topic")
seq = 1
while True:
    time.sleep(0.05)
    letter = random.choice("abcdefghijklmnopqrstuvwxyz")
    topic = "msg.topic.%s" % letter
    bus.send(topic, "message %d" % seq)
    seq += 1
    print("", letter, seq, end="\r")



from __future__ import print_function
import os

import Pyro4


@Pyro4.expose
class Thingy(object):
    def message(self, arg):
        print("Message received:", arg)
        return "Roger!"


if os.path.exists("example_unix.sock"):
    os.remove("example_unix.sock")

with Pyro4.Daemon(unixsocket="example_unix.sock") as d:
    uri = d.register(Thingy, "example.unixsock")
    print("Server running, uri=", uri)
    d.requestLoop()



from __future__ import print_function
import sys

import Pyro4


if sys.version_info < (3, 0):
    input = raw_input

with Pyro4.core.Proxy("PYRONAME:example.embedded.server") as proxy:
    print("5*11=%d" % proxy.multiply(5, 11))
    print("'x'*10=%s" % proxy.multiply('x', 10))

    input("press enter to do a loop of some more calls:")
    for i in range(1, 20):
        print("2*i=%d" % proxy.multiply(2, i))
        print("'@'*i=%s" % proxy.multiply('@', i))



import Pyro4


# Chat box administration server.
# Handles logins, logouts, channels and nicknames, and the chatting.
@Pyro4.expose
@Pyro4.behavior(instance_mode="single")
class ChatBox(object):
    def __init__(self):
        self.channels = {}  # registered channels { channel --> (nick, client callback) list }
        self.nicks = []  # all registered nicks on this server

    def getChannels(self):
        return list(self.channels.keys())

    def getNicks(self):
        return self.nicks

    def join(self, channel, nick, callback):
        if not channel or not nick:
            raise ValueError("invalid channel or nick name")
        if nick in self.nicks:
            raise ValueError('this nick is already in use')
        if channel not in self.channels:
            print('CREATING NEW CHANNEL %s' % channel)
            self.channels[channel] = []
        self.channels[channel].append((nick, callback))
        self.nicks.append(nick)
        print("%s JOINED %s" % (nick, channel))
        self.publish(channel, 'SERVER', '** ' + nick + ' joined **')
        return [nick for (nick, c) in self.channels[channel]]  # return all nicks in this channel

    def leave(self, channel, nick):
        if channel not in self.channels:
            print('IGNORED UNKNOWN CHANNEL %s' % channel)
            return
        for (n, c) in self.channels[channel]:
            if n == nick:
                self.channels[channel].remove((n, c))
                break
        self.publish(channel, 'SERVER', '** ' + nick + ' left **')
        if len(self.channels[channel]) < 1:
            del self.channels[channel]
            print('REMOVED CHANNEL %s' % channel)
        self.nicks.remove(nick)
        print("%s LEFT %s" % (nick, channel))

    def publish(self, channel, nick, msg):
        if channel not in self.channels:
            print('IGNORED UNKNOWN CHANNEL %s' % channel)
            return
        for (n, c) in self.channels[channel][:]:  # use a copy of the list
            try:
                c.message(nick, msg)  # oneway call
            except Pyro4.errors.ConnectionClosedError:
                # connection dropped, remove the listener if it's still there
                # check for existence because other thread may have killed it already
                if (n, c) in self.channels[channel]:
                    self.channels[channel].remove((n, c))
                    print('Removed dead listener %s %s' % (n, c))


Pyro4.Daemon.serveSimple({
    ChatBox: "example.chatbox.server"
})



import Pyro4
import Pyro4.errors
from diffiehellman import DiffieHellman

Pyro4.config.SERVERTYPE = "multiplex"


@Pyro4.expose
class SecretStuff(object):
    def process(self, message):
        if not self._pyroDaemon._pyroHmacKey:
            raise Pyro4.errors.PyroError("no hmac key has been set, can't call this method!")
        print("Got the message:", message)
        return "message was received ok"

ns = Pyro4.locateNS()
daemon = Pyro4.Daemon()
daemon._pyroHmacKey = b"will be set to shared secret key by KeyExchange"
uri = daemon.register(SecretStuff)
ns.register("example.dh.secretstuff", uri)


@Pyro4.behavior(instance_mode="session")
class KeyExchange(object):
    def __init__(self):
        print("New KeyExchange, initializing Diffie-Hellman")
        self.dh = DiffieHellman(group=14)

    @Pyro4.expose
    def exchange_key(self, other_public_key):
        print("received a public key, calculating shared secret...")
        self.dh.make_shared_secret_and_key(other_public_key)
        print("setting new shared secret key.")
        global daemon
        daemon._pyroHmacKey = self.dh.key
        return self.dh.public_key


# The key exchange service can't be part of the same daemon as the
# other service because it must not have a Hmac key set on the daemon.
# So we create another daemon without hmac key and combine it.
key_daemon = Pyro4.Daemon()
uri = key_daemon.register(KeyExchange)
ns.register("example.dh.keyexchange", uri)
ns._pyroRelease()

key_daemon.combine(daemon)
print("Starting server loop...")
key_daemon.requestLoop()



"""
Pyro MessageBus:  a simple pub/sub message bus.
Provides a way of communicating where the sender and receivers are fully decoupled.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

PYRO_MSGBUS_NAME = "Pyro.MessageBus"



"""
This is the subscriber for the 'many topics' messages example.
For code with more explanations, see the regular 'weather' message example code.
"""
from __future__ import print_function
import os
import time
import threading
import Pyro4
from operator import itemgetter
from messagebus.messagebus import Subscriber

Pyro4.config.AUTOPROXY = True


@Pyro4.expose
class Subber(Subscriber):
    def init_counters(self, topics):
        self.message_counter = {}
        self.last_message = {}
        for t in topics:
            self.message_counter[t] = 0
            self.last_message[t] = None

    def consume_message(self, topic, message):
        self.message_counter[topic] += 1
        self.last_message[topic] = message


def clear_screen():
    os.system(['clear', 'cls'][os.name == 'nt'])


subber = Subber()
d = Pyro4.Daemon()
d.register(subber)
daemon_thread = threading.Thread(target=d.requestLoop)
daemon_thread.daemon = True
daemon_thread.start()

# mass subscribe to all available topics
topics = list(sorted(subber.bus.topics()))
subber.init_counters(topics)
for t in topics:
    subber.bus.subscribe(t, subber)

# show a table of the active topics on the bus
while True:
    clear_screen()
    print(time.ctime(), "-- active topics on the messagebus:")
    print("{:20} : {:5}     {}         {}".format("topic", "count", "last_recv", "last message data"))
    for topic, count in sorted(subber.message_counter.items(), key=itemgetter(1), reverse=True):
        msg = subber.last_message[topic]
        if msg:
            print("{:20} : {:5d}  -  {}   {!r:.20}".format(topic, count, msg.created.time(), msg.data))
        else:
            print("{:20} : {:5d}".format(topic, count))
    print("(restart me to refresh the list of topics)")
    time.sleep(1)



from __future__ import print_function
import sys
import Pyro4
import uuid

# example: set a single correlation id on the context that should be passed along
Pyro4.current_context.correlation_id = uuid.uuid4()
print("correlation id set to:", Pyro4.current_context.correlation_id)

if sys.version_info < (3, 0):
    input = raw_input


# custom proxy needed to get to annotation data, before Pyro 4.56
class CustomAnnotationProxy(Pyro4.Proxy):
    def __init__(self, uri):
        super(CustomAnnotationProxy, self).__init__(uri)
        self._pyroHmacKey = b"secr3t_k3y"

    # override the method that adds annotations and add our own
    def _pyroAnnotations(self):
        return {"XYZZ": b"Hello, I am a custom annotation from the proxy!"}

    def _pyroResponseAnnotations(self, annotations, msgtype):
        print("    Got response (msgtype=%d). Annotations:" % msgtype)
        for key in annotations:
            if key == "CORR":
                value = uuid.UUID(bytes=annotations[key])
            elif key == "HMAC":
                value = "[...]"
            else:
                value = annotations[key]
            print("      {0} -> {1}".format(key, value))


uri = input("Enter the URI of the server object: ")

print("\n------- (older) method to get annotations via callback on custom proxy... -----\n")
with CustomAnnotationProxy(uri) as proxy:
    print("Sending a few messages using one proxy...")
    for i in range(4):
        msg = proxy.echo("hello-%d" % i)

proxies = [CustomAnnotationProxy(uri) for _ in range(5)]
for p in proxies:
    print("Sending one message from new proxy...")
    msg = p.echo("hello-%d" % id(p))
    p._pyroRelease()

with CustomAnnotationProxy(uri) as proxy:
    # oneway
    print("Sending a oneway message... (should only print a connection ok response)")
    proxy.oneway("hello-ONEWAY-1")
    print("Sending another oneway message... (should not print a response at all)")
    proxy.oneway("hello-ONEWAY-2")
    # asynchronous
    print("Asynchronous proxy message...")
    proxy._pyroAsync()
    result = proxy.echo("hello-ASYNC")
    _ = result.value


print("\n------- get annotations via normal proxy and the call context... -----\n")
input("press enter:")
# the code below works as of Pyro 4.56.
with Pyro4.Proxy(uri) as proxy:
    proxy._pyroHmacKey = b"secr3t_k3y"
    print("normal call")

    Pyro4.current_context.annotations = {"XYZZ": b"custom annotation from client via new way(1)"}
    result = proxy.echo("hi there - new method of annotation access in client")
    print("Annotations in response were: ", Pyro4.current_context.response_annotations)

    print("\noneway call")
    Pyro4.current_context.annotations = {"XYZZ": b"custom annotation from client via new way(2)"}
    proxy.oneway("hi there ONEWAY - new method of annotation access in client")
    print("Annotations in response were: ", Pyro4.current_context.response_annotations)
    print("   (should be empty because oneway!)")


print("\nSee the console output on the server for more results.")



import random
import time


class StockMarket(object):
    def __init__(self, marketname, symbols):
        self.name = marketname
        self.symbols = symbols

    def quotes(self):
        while True:
            symbol = random.choice(self.symbols)
            yield symbol, round(random.uniform(5, 150), 2)
            time.sleep(random.random()/2.0)



import Pyro4


@Pyro4.expose
class bench(object):
    def length(self, string):
        return len(string)

    def timestwo(self, value):
        return value * 2

    def bigreply(self):
        return 'BIG REPLY' * 500

    def bigarg(self, arg):
        return len(arg)

    def manyargs(self, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15):
        return a1 + a2 + a3 + a4 + a5 + a6 + a7 + a8 + a9 + a10 + a11 + a12 + a13 + a14 + a15

    def noreply(self, arg):
        pass

    def varargs(self, *args):
        return len(args)

    def keywords(self, **args):
        return args

    def echo(self, *args):
        return args

    @Pyro4.oneway
    def oneway(self, *args):
        # oneway doesn't return anything
        pass



"""
Tests for the core logic.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

from __future__ import print_function
import copy
import logging
import os
import sys
import time
import uuid
import socket
import unittest
import warnings
import Pyro4.core
import Pyro4.errors
import Pyro4.constants
import Pyro4.futures
from Pyro4.configuration import config
from testsupport import *


if (3, 0) <= sys.version_info < (3, 4):
    import imp
    reload = imp.reload
elif sys.version_info >= (3, 4):
    import importlib
    reload = importlib.reload


class CoreTests(unittest.TestCase):

    def setUp(self):
        warnings.filterwarnings("ignore")

    def tearDown(self):
        warnings.resetwarnings()

    def testProxyNoHmac(self):
        # check that proxy without hmac is possible
        with Pyro4.core.Proxy("PYRO:object@host:9999") as p:
            pass

    def testDaemonNoHmac(self):
        # check that daemon without hmac is possible
        d = Pyro4.core.Daemon()
        d.shutdown()

    def testConfig(self):
        self.assertTrue(type(config.COMPRESSION) is bool)
        self.assertTrue(type(config.NS_PORT) is int)
        cfgdict = config.asDict()
        self.assertTrue(type(cfgdict) is dict)
        self.assertIn("COMPRESSION", cfgdict)
        self.assertEqual(config.COMPRESSION, cfgdict["COMPRESSION"])

    def testConfigDefaults(self):
        # some security sensitive settings:
        config.reset(False)   # reset the config to default
        self.assertTrue(config.REQUIRE_EXPOSE)
        self.assertEqual("localhost", config.HOST)
        self.assertEqual("localhost", config.NS_HOST)
        self.assertFalse(config.FLAME_ENABLED)
        self.assertEqual("serpent", config.SERIALIZER)
        self.assertEqual({"json", "serpent", "marshal"}, config.SERIALIZERS_ACCEPTED)

    def testConfigValid(self):
        try:
            config.XYZ_FOOBAR = True  # don't want to allow weird config names
            self.fail("expected exception for weird config item")
        except AttributeError:
            pass

    def testConfigParseBool(self):
        config = Pyro4.configuration.Configuration()
        self.assertTrue(type(config.COMPRESSION) is bool)
        os.environ["PYRO_COMPRESSION"] = "yes"
        config.reset()
        self.assertTrue(config.COMPRESSION)
        os.environ["PYRO_COMPRESSION"] = "off"
        config.reset()
        self.assertFalse(config.COMPRESSION)
        os.environ["PYRO_COMPRESSION"] = "foobar"
        self.assertRaises(ValueError, config.reset)
        del os.environ["PYRO_COMPRESSION"]
        config.reset()

    def testConfigDump(self):
        config = Pyro4.configuration.Configuration()
        dump = config.dump()
        self.assertIn("version:", dump)
        self.assertIn("LOGLEVEL", dump)

    def testLogInit(self):
        _ = logging.getLogger("Pyro4")
        os.environ["PYRO_LOGLEVEL"] = "DEBUG"
        os.environ["PYRO_LOGFILE"] = "{stderr}"
        reload(Pyro4)
        _ = logging.getLogger("Pyro4")
        os.environ["PYRO_LOGFILE"] = "Pyro.log"
        reload(Pyro4)
        _ = logging.getLogger("Pyro4")
        del os.environ["PYRO_LOGLEVEL"]
        del os.environ["PYRO_LOGFILE"]
        reload(Pyro4)
        _ = logging.getLogger("Pyro4")

    def testUriStrAndRepr(self):
        uri = "PYRONAME:some_obj_name"
        p = Pyro4.core.URI(uri)
        self.assertEqual(uri, str(p))
        uri = "PYRONAME:some_obj_name@host.com"
        p = Pyro4.core.URI(uri)
        self.assertEqual(uri + ":" + str(config.NS_PORT), str(p))  # a PYRONAME uri with a hostname gets a port too if omitted
        uri = "PYRONAME:some_obj_name@host.com:8888"
        p = Pyro4.core.URI(uri)
        self.assertEqual(uri, str(p))
        expected = "<Pyro4.core.URI at 0x%x; PYRONAME:some_obj_name@host.com:8888>" % id(p)
        self.assertEqual(expected, repr(p))
        uri = "PYRO:12345@host.com:9999"
        p = Pyro4.core.URI(uri)
        self.assertEqual(uri, str(p))
        self.assertEqual(uri, p.asString())
        uri = "PYRO:12345@./u:sockname"
        p = Pyro4.core.URI(uri)
        self.assertEqual(uri, str(p))
        uri = "PYRO:12345@./u:sockname"
        unicodeuri = unicode(uri)
        p = Pyro4.core.URI(unicodeuri)
        self.assertEqual(uri, str(p))
        self.assertEqual(unicodeuri, unicode(p))
        self.assertTrue(type(p.sockname) is unicode)
        uri = "PYRO:12345@./u:sock name with strings"
        p = Pyro4.core.URI(uri)
        self.assertEqual(uri, str(p))

    def testUriParsingPyro(self):
        p = Pyro4.core.URI("PYRONAME:some_obj_name")
        self.assertEqual("PYRONAME", p.protocol)
        self.assertEqual("some_obj_name", p.object)
        self.assertIsNone(p.host)
        self.assertIsNone(p.sockname)
        self.assertIsNone(p.port)
        p = Pyro4.core.URI("PYRONAME:some_obj_name@host.com:9999")
        self.assertEqual("PYRONAME", p.protocol)
        self.assertEqual("some_obj_name", p.object)
        self.assertEqual("host.com", p.host)
        self.assertEqual(9999, p.port)

        p = Pyro4.core.URI("PYRO:12345@host.com:4444")
        self.assertEqual("PYRO", p.protocol)
        self.assertEqual("12345", p.object)
        self.assertEqual("host.com", p.host)
        self.assertIsNone(p.sockname)
        self.assertEqual(4444, p.port)
        self.assertEqual("host.com:4444", p.location)
        p = Pyro4.core.URI("PYRO:12345@./u:sockname")
        self.assertEqual("12345", p.object)
        self.assertEqual("sockname", p.sockname)
        p = Pyro4.core.URI("PYRO:12345@./u:/tmp/sockname")
        self.assertEqual("12345", p.object)
        self.assertEqual("/tmp/sockname", p.sockname)
        p = Pyro4.core.URI("PYRO:12345@./u:/path with spaces/sockname  ")
        self.assertEqual("12345", p.object)
        self.assertEqual("/path with spaces/sockname  ", p.sockname)
        p = Pyro4.core.URI("PYRO:12345@./u:../sockname")
        self.assertEqual("12345", p.object)
        self.assertEqual("../sockname", p.sockname)
        p = Pyro4.core.URI("pyro:12345@host.com:4444")
        self.assertEqual("PYRO", p.protocol)
        self.assertEqual("12345", p.object)
        self.assertEqual("host.com", p.host)
        self.assertIsNone(p.sockname)
        self.assertEqual(4444, p.port)

    def testUriParsingIpv6(self):
        p = Pyro4.core.URI("pyro:12345@[::1]:4444")
        self.assertEqual("::1", p.host)
        self.assertEqual("[::1]:4444", p.location)
        with self.assertRaises(Pyro4.errors.PyroError) as e:
            Pyro4.core.URI("pyro:12345@[[::1]]:4444")
        self.assertEqual("invalid ipv6 address: enclosed in too many brackets", str(e.exception))
        with self.assertRaises(Pyro4.errors.PyroError) as e:
            Pyro4.core.URI("pyro:12345@[must_be_numeric_here]:4444")
        self.assertEqual("invalid ipv6 address: the part between brackets must be a numeric ipv6 address", str(e.exception))

    def testUriParsingPyroname(self):
        p = Pyro4.core.URI("PYRONAME:objectname")
        self.assertEqual("PYRONAME", p.protocol)
        self.assertEqual("objectname", p.object)
        self.assertIsNone(p.host)
        self.assertIsNone(p.port)
        p = Pyro4.core.URI("PYRONAME:objectname@nameserverhost")
        self.assertEqual("PYRONAME", p.protocol)
        self.assertEqual("objectname", p.object)
        self.assertEqual("nameserverhost", p.host)
        self.assertEqual(config.NS_PORT, p.port)  # Pyroname uri with host gets a port too if not specified
        p = Pyro4.core.URI("PYRONAME:objectname@nameserverhost:4444")
        self.assertEqual("PYRONAME", p.protocol)
        self.assertEqual("objectname", p.object)
        self.assertEqual("nameserverhost", p.host)
        self.assertEqual(4444, p.port)
        p = Pyro4.core.URI("PyroName:some_obj_name@host.com:9999")
        self.assertEqual("PYRONAME", p.protocol)
        p = Pyro4.core.URI("pyroname:some_obj_name@host.com:9999")
        self.assertEqual("PYRONAME", p.protocol)

    def testUriParsingPyrometa(self):
        p = Pyro4.core.URI("PYROMETA:meta")
        self.assertEqual("PYROMETA", p.protocol)
        self.assertEqual({"meta"}, p.object)
        self.assertIsNone(p.host)
        self.assertIsNone(p.port)
        p = Pyro4.core.URI("PYROMETA:meta1,meta2,meta2@nameserverhost")
        self.assertEqual("PYROMETA", p.protocol)
        self.assertEqual({"meta1", "meta2"}, p.object)
        self.assertEqual("nameserverhost", p.host)
        self.assertEqual(config.NS_PORT, p.port)  # PyroMeta uri with host gets a port too if not specified
        p = Pyro4.core.URI("PYROMETA:meta@nameserverhost:4444")
        self.assertEqual("PYROMETA", p.protocol)
        self.assertEqual({"meta"}, p.object)
        self.assertEqual("nameserverhost", p.host)
        self.assertEqual(4444, p.port)
        p = Pyro4.core.URI("PyroMeta:meta1,meta2@host.com:9999")
        self.assertEqual("PYROMETA", p.protocol)
        p = Pyro4.core.URI("PyroMeta:meta1,meta2@host.com:9999")
        self.assertEqual("PYROMETA", p.protocol)

    def testInvalidUris(self):
        self.assertRaises(TypeError, Pyro4.core.URI, None)
        self.assertRaises(TypeError, Pyro4.core.URI, 99999)
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, " ")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "a")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYR")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO::")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:a")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:x@")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:x@hostname")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:@hostname:portstr")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:@hostname:7766")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:objid@hostname:7766:bogus")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:obj id@hostname:7766")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYROLOC:objname")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYROLOC:objname@host")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYROLOC:objectname@hostname:4444")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRONAME:")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRONAME:obj name@nameserver:bogus")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRONAME:objname@nameserver:bogus")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRONAME:objname@nameserver:7766:bogus")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYROMETA:")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYROMETA:meta@nameserver:bogus")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYROMETA:meta@nameserver:7766:bogus")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYROMETA:meta1, m2 ,m3@nameserver:7766:bogus")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "FOOBAR:")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "FOOBAR:objid@hostname:7766")
        self.assertRaises(Pyro4.errors.PyroError, Pyro4.core.URI, "PYRO:12345@./u:sockname:9999")

    def testUriUnicode(self):
        p = Pyro4.core.URI(unicode("PYRO:12345@host.com:4444"))
        self.assertEqual("PYRO", p.protocol)
        self.assertEqual("12345", p.object)
        self.assertEqual("host.com", p.host)
        self.assertTrue(type(p.protocol) is unicode)
        self.assertTrue(type(p.object) is unicode)
        self.assertTrue(type(p.host) is unicode)
        self.assertIsNone(p.sockname)
        self.assertEqual(4444, p.port)

        uri = "PYRO:12345@hostname:9999"
        p = Pyro4.core.URI(uri)
        pu = Pyro4.core.URI(unicode(uri))
        self.assertEqual("PYRO", pu.protocol)
        self.assertEqual("hostname", pu.host)
        self.assertEqual(p, pu)
        self.assertEqual(str(p), str(pu))
        unicodeuri = "PYRO:weirdchars" + unichr(0x20ac) + "@host" + unichr(0x20AC) + ".com:4444"
        pu = Pyro4.core.URI(unicodeuri)
        self.assertEqual("PYRO", pu.protocol)
        self.assertEqual("host" + unichr(0x20AC) + ".com", pu.host)
        self.assertEqual("weirdchars" + unichr(0x20AC), pu.object)
        if sys.version_info <= (3, 0):
            self.assertEqual("PYRO:weirdchars?@host?.com:4444", pu.__str__())
            expected = "<Pyro4.core.URI at 0x%x; PYRO:weirdchars?@host?.com:4444>" % id(pu)
            self.assertEqual(expected, repr(pu))
        else:
            self.assertEqual("PYRO:weirdchars" + unichr(0x20ac) + "@host" + unichr(0x20ac) + ".com:4444", pu.__str__())
            expected = ("<Pyro4.core.URI at 0x%x; PYRO:weirdchars" + unichr(0x20ac) + "@host" + unichr(0x20ac) + ".com:4444>") % id(pu)
            self.assertEqual(expected, repr(pu))
        self.assertEqual("PYRO:weirdchars" + unichr(0x20ac) + "@host" + unichr(0x20ac) + ".com:4444", pu.asString())
        self.assertEqual("PYRO:weirdchars" + unichr(0x20ac) + "@host" + unichr(0x20ac) + ".com:4444", unicode(pu))

    def testUriCopy(self):
        p1 = Pyro4.core.URI("PYRO:12345@hostname:9999")
        p2 = Pyro4.core.URI(p1)
        p3 = copy.copy(p1)
        self.assertEqual(p1.protocol, p2.protocol)
        self.assertEqual(p1.host, p2.host)
        self.assertEqual(p1.port, p2.port)
        self.assertEqual(p1.object, p2.object)
        self.assertEqual(p1, p2)
        self.assertEqual(p1.protocol, p3.protocol)
        self.assertEqual(p1.host, p3.host)
        self.assertEqual(p1.port, p3.port)
        self.assertEqual(p1.object, p3.object)
        self.assertEqual(p1, p3)

    def testUriSubclassCopy(self):
        class SubURI(Pyro4.core.URI):
            pass
        u = SubURI("PYRO:12345@hostname:9999")
        u2 = copy.copy(u)
        self.assertIsInstance(u2, SubURI)

    def testUriEqual(self):
        p1 = Pyro4.core.URI("PYRO:12345@host.com:9999")
        p2 = Pyro4.core.URI("PYRO:12345@host.com:9999")
        p3 = Pyro4.core.URI("PYRO:99999@host.com:4444")
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)
        self.assertNotEqual(p2, p3)
        self.assertTrue(p1 == p2)
        self.assertFalse(p1 == p3)
        self.assertFalse(p2 == p3)
        self.assertFalse(p1 != p2)
        self.assertTrue(p1 != p3)
        self.assertTrue(p2 != p3)
        self.assertTrue(hash(p1) == hash(p2))
        self.assertTrue(hash(p1) != hash(p3))
        p2.port = 4444
        p2.object = "99999"
        self.assertNotEqual(p1, p2)
        self.assertEqual(p2, p3)
        self.assertFalse(p1 == p2)
        self.assertTrue(p2 == p3)
        self.assertTrue(p1 != p2)
        self.assertFalse(p2 != p3)
        self.assertTrue(hash(p1) != hash(p2))
        self.assertTrue(hash(p2) == hash(p3))
        self.assertFalse(p1 == 42)
        self.assertTrue(p1 != 42)

    def testLocation(self):
        self.assertTrue(Pyro4.core.URI.isUnixsockLocation("./u:name"))
        self.assertFalse(Pyro4.core.URI.isUnixsockLocation("./p:name"))
        self.assertFalse(Pyro4.core.URI.isUnixsockLocation("./x:name"))
        self.assertFalse(Pyro4.core.URI.isUnixsockLocation("foobar"))

    def testProxyConnectedSocket(self):
        s1, s2 = socket.socketpair()
        s1.settimeout(0.1)
        s2.settimeout(0.1)
        try:
            Pyro4.config.METADATA = False
            with self.assertRaises(Pyro4.errors.PyroError):
                Pyro4.Proxy("test")
            with Pyro4.Proxy("test", connected_socket=s1) as p:
                self.assertEquals("<<connected-socket>>:0", p._pyroUri.location)
                self.assertIsNotNone(p._pyroConnection)
                p._pyroRelease()
                self.assertIsNotNone(p._pyroConnection)
            self.assertIsNotNone(p._pyroConnection)
        finally:
            Pyro4.config.METADATA = True

    def testProxyCopy(self):
        u = Pyro4.core.URI("PYRO:12345@hostname:9999")
        p1 = Pyro4.core.Proxy(u)
        p2 = copy.copy(p1)  # check that most basic copy also works
        self.assertEqual(p1, p2)
        self.assertEqual(set(), p2._pyroOneway)
        p1._pyroAttrs = set("abc")
        p1._pyroTimeout = 42
        p1._pyroOneway = set("def")
        p1._pyroMethods = set("ghi")
        p1._pyroHmacKey = b"secret"
        p1._pyroHandshake = "apples"
        p2 = copy.copy(p1)
        self.assertEqual(p1, p2)
        self.assertEqual(p1._pyroUri, p2._pyroUri)
        self.assertEqual(p1._pyroOneway, p2._pyroOneway)
        self.assertEqual(p1._pyroMethods, p2._pyroMethods)
        self.assertEqual(p1._pyroAttrs, p2._pyroAttrs)
        self.assertEqual(p1._pyroTimeout, p2._pyroTimeout)
        self.assertEqual(p1._pyroHmacKey, p2._pyroHmacKey)
        self.assertEqual(p1._pyroHandshake, p2._pyroHandshake)
        p1._pyroRelease()
        p2._pyroRelease()

    def testProxySubclassCopy(self):
        class ProxySub(Pyro4.core.Proxy):
            pass
        p = ProxySub("PYRO:12345@hostname:9999")
        p2 = copy.copy(p)
        self.assertIsInstance(p2, ProxySub)
        p._pyroRelease()
        p2._pyroRelease()

    def testAsyncProxyCopy(self):
        try:
            config.METADATA = False
            with Pyro4.core.Proxy("PYRO:12345@hostname:9999") as proxy:
                proxy._pyroAsync()
                p2 = copy.copy(proxy)
                asynccall = p2.foobar()
                self.assertIsInstance(asynccall, Pyro4.futures.FutureResult)
        finally:
            config.METADATA = True

    def testBatchProxyAdapterCopy(self):
        with Pyro4.core.Proxy("PYRO:12345@hostname:9999") as proxy:
            batchproxy = proxy._pyroBatch()
            p2 = copy.copy(batchproxy)
            self.assertIsInstance(p2, Pyro4.core._BatchProxyAdapter)

    def testProxyOffline(self):
        # only offline stuff here.
        # online stuff needs a running daemon, so we do that in another test, to keep this one simple
        self.assertRaises(TypeError, Pyro4.core.Proxy, 999)  # wrong arg
        p1 = Pyro4.core.Proxy("PYRO:9999@localhost:15555")
        p2 = Pyro4.core.Proxy(Pyro4.core.URI("PYRO:9999@localhost:15555"))
        self.assertEqual(p1._pyroUri, p2._pyroUri)
        self.assertIsNone(p1._pyroConnection)
        p1._pyroRelease()
        p1._pyroRelease()
        # try copying a not-connected proxy
        p3 = copy.copy(p1)
        self.assertIsNone(p3._pyroConnection)
        self.assertIsNone(p1._pyroConnection)
        self.assertEqual(p3._pyroUri, p1._pyroUri)
        self.assertIsNot(p3._pyroUri, p1._pyroUri)
        p3._pyroRelease()

    def testProxyRepr(self):
        with Pyro4.core.Proxy("PYRO:9999@localhost:15555") as p:
            address = id(p)
            expected = "<Pyro4.core.Proxy at 0x%x; not connected; for PYRO:9999@localhost:15555>" % address
            self.assertEqual(expected, repr(p))
            self.assertEqual(unicode(expected), unicode(p))

    def testProxySerializerOverride(self):
        serializer = config.SERIALIZER
        try:
            config.SERIALIZER = "~invalid~"
            _ = Pyro4.core.Proxy("PYRO:obj@localhost:5555")
            self.fail("must raise exception")
        except Pyro4.errors.SerializeError as x:
            self.assertIn("~invalid~", str(x))
            self.assertIn("unknown", str(x))
        finally:
            config.SERIALIZER = serializer
        try:
            proxy = Pyro4.core.Proxy("PYRO:obj@localhost:5555")
            proxy._pyroSerializer = "~invalidoverride~"
            proxy._pyroConnection = "FAKE"
            proxy.methodcall()
            self.fail("must raise exception")
        except Pyro4.errors.SerializeError as x:
            self.assertIn("~invalidoverride~", str(x))
            self.assertIn("unknown", str(x))
        finally:
            proxy._pyroConnection = None
            config.SERIALIZER = serializer

    def testProxyDir(self):
        # PyPy tries to call deprecated __members__ and __methods__
        # that causes a CommunicationError since we use a fake URI
        class ProxyWithFixedGettattr(Pyro4.core.Proxy):
            def __getattr__(self, name):
                if name in ('__members__', '__methods__'):
                    raise AttributeError(name)
                return super(Pyro4.core.Proxy, self).__getattr__(name)
        ProxyClass = ProxyWithFixedGettattr if 'pypy' in sys.version.lower() else Pyro4.core.Proxy
        p = ProxyClass("PYRO:9999@localhost:15555")
        # make sure that __dir__ implementation works the same way as dir()
        dir_result = dir(p)
        if sys.version_info < (3, 3):
            # before 3.3 python's object class didn't have __dir__ method
            dir_result.remove('__dir__')
        old_dir_method = getattr(Pyro4.core.Proxy, '__dir__')
        try:
            delattr(Pyro4.core.Proxy, '__dir__')
            self.assertEqual(dir_result, dir(p))
        finally:
            setattr(Pyro4.core.Proxy, '__dir__', old_dir_method)
        p._pyroRelease()

    def testProxyDirMetadata(self):
        with Pyro4.core.Proxy("PYRO:9999@localhost:15555") as p:
            # metadata isn't loaded
            self.assertIn('__hash__', dir(p))
            self.assertNotIn('ping', dir(p))
            # emulate obtaining metadata
            p._pyroAttrs = {"prop"}
            p._pyroMethods = {"ping"}
            self.assertIn('__hash__', dir(p))
            self.assertIn('prop', dir(p))
            self.assertIn('ping', dir(p))

    def testProxySettings(self):
        p1 = Pyro4.core.Proxy("PYRO:9999@localhost:15555")
        p2 = Pyro4.core.Proxy("PYRO:9999@localhost:15555")
        p1._pyroOneway.add("method")
        p1._pyroAttrs.add("attr")
        p1._pyroMethods.add("method2")
        self.assertIn("method", p1._pyroOneway)
        self.assertIn("attr", p1._pyroAttrs)
        self.assertIn("method2", p1._pyroMethods)
        self.assertNotIn("method", p2._pyroOneway)
        self.assertNotIn("attr", p2._pyroAttrs)
        self.assertNotIn("method2", p2._pyroMethods)
        self.assertIsNot(p1._pyroOneway, p2._pyroOneway, "p1 and p2 should have different oneway tables")
        self.assertIsNot(p1._pyroAttrs, p2._pyroAttrs, "p1 and p2 should have different attr tables")
        self.assertIsNot(p1._pyroMethods, p2._pyroMethods, "p1 and p2 should have different method tables")
        p1._pyroRelease()
        p2._pyroRelease()

    def testProxyWithStmt(self):
        class ConnectionMock(object):
            closeCalled = False
            keep_open = False

            def close(self):
                self.closeCalled = True

        connMock = ConnectionMock()
        # first without a 'with' statement
        p = Pyro4.core.Proxy("PYRO:9999@localhost:15555")
        p._pyroConnection = connMock
        self.assertFalse(connMock.closeCalled)
        p._pyroRelease()
        self.assertIsNone(p._pyroConnection)
        self.assertTrue(connMock.closeCalled)

        connMock = ConnectionMock()
        with Pyro4.core.Proxy("PYRO:9999@localhost:15555") as p:
            p._pyroConnection = connMock
        self.assertIsNone(p._pyroConnection)
        self.assertTrue(connMock.closeCalled)
        connMock = ConnectionMock()
        try:
            with Pyro4.core.Proxy("PYRO:9999@localhost:15555") as p:
                p._pyroConnection = connMock
                print(1 // 0)  # cause an error
            self.fail("expected error")
        except ZeroDivisionError:
            pass
        self.assertIsNone(p._pyroConnection)
        self.assertTrue(connMock.closeCalled)
        p = Pyro4.core.Proxy("PYRO:9999@localhost:15555")
        with p:
            self.assertIsNotNone(p._pyroUri)
        with p:
            self.assertIsNotNone(p._pyroUri)
        p._pyroRelease()

    def testProxyHmac(self):
        class ConnectionMock(object):
            def __init__(self):
                self.msgbytes = None
                self.keep_open = False
            def close(self):
                pass
            def send(self, msgbytes):
                self.msgbytes = msgbytes
            def recv(self, size):
                raise Pyro4.errors.ConnectionClosedError("mock")
        proxy = Pyro4.core.Proxy("PYRO:foobar@localhost:59999")
        proxy._pyroHmacKey = b"secret"
        conn_mock = ConnectionMock()
        proxy._pyroConnection = conn_mock
        with self.assertRaises(Pyro4.errors.ConnectionClosedError):
            proxy.foo()
        self.assertIn(b"HMAC", conn_mock.msgbytes)

        conn_mock = ConnectionMock()
        proxy._pyroConnection = conn_mock
        proxy._pyroHmacKey = None
        with self.assertRaises(Pyro4.errors.ConnectionClosedError):
            proxy.foo()
        self.assertNotIn(b"HMAC", conn_mock.msgbytes)
        self.assertIsNone(proxy._pyroHmacKey)
        proxy._pyroRelease()

    def testNoConnect(self):
        wrongUri = Pyro4.core.URI("PYRO:foobar@localhost:59999")
        with Pyro4.core.Proxy(wrongUri) as p:
            try:
                p.ping()
                self.fail("CommunicationError expected")
            except Pyro4.errors.CommunicationError:
                pass

    def testTimeoutGetSet(self):
        class ConnectionMock(object):
            def __init__(self):
                self.timeout = config.COMMTIMEOUT
                self.keep_open = False

            def close(self):
                pass

        config.COMMTIMEOUT = None
        p = Pyro4.core.Proxy("PYRO:obj@host:555")
        self.assertIsNone(p._pyroTimeout)
        p._pyroTimeout = 5
        self.assertEqual(5, p._pyroTimeout)
        p = Pyro4.core.Proxy("PYRO:obj@host:555")
        p._pyroConnection = ConnectionMock()
        self.assertIsNone(p._pyroTimeout)
        p._pyroTimeout = 5
        self.assertEqual(5, p._pyroTimeout)
        self.assertEqual(5, p._pyroConnection.timeout)
        config.COMMTIMEOUT = 2
        p = Pyro4.core.Proxy("PYRO:obj@host:555")
        p._pyroConnection = ConnectionMock()
        self.assertEqual(2, p._pyroTimeout)
        self.assertEqual(2, p._pyroConnection.timeout)
        p._pyroTimeout = None
        self.assertIsNone(p._pyroTimeout)
        self.assertIsNone(p._pyroConnection.timeout)
        config.COMMTIMEOUT = None
        p._pyroRelease()

    def testCallbackDecorator(self):
        # just test the decorator itself, testing the callback
        # exception handling is kinda hard in unit tests. Maybe later.
        class Test(object):
            @Pyro4.core.callback
            def method(self):
                pass

            def method2(self):
                pass

        t = Test()
        self.assertEqual(True, getattr(t.method, "_pyroCallback"))
        self.assertEqual(False, getattr(t.method2, "_pyroCallback", False))

    def testProxyEquality(self):
        p1 = Pyro4.core.Proxy("PYRO:thing@localhost:15555")
        p2 = Pyro4.core.Proxy("PYRO:thing@localhost:15555")
        p3 = Pyro4.core.Proxy("PYRO:other@machine:16666")
        self.assertTrue(p1 == p2)
        self.assertFalse(p1 != p2)
        self.assertFalse(p1 == p3)
        self.assertTrue(p1 != p3)
        self.assertTrue(hash(p1) == hash(p2))
        self.assertFalse(hash(p1) == hash(p3))
        self.assertFalse(p1 == 42)
        self.assertTrue(p1 != 42)
        p1._pyroRelease()
        p2._pyroRelease()
        p3._pyroRelease()


    def testCallContext(self):
        ctx = Pyro4.core.current_context
        corr_id = uuid.UUID('1897022f-c481-4117-a4cc-cbd1ca100582')
        ctx.correlation_id = corr_id
        d = ctx.to_global()
        self.assertIsInstance(d, dict)
        self.assertEqual(corr_id, d["correlation_id"])
        corr_id2 = uuid.UUID('67b05ad9-2d6a-4ed8-8ed5-95cba68b4cf9')
        d["correlation_id"] = corr_id2
        ctx.from_global(d)
        self.assertEqual(corr_id2, Pyro4.core.current_context.correlation_id)
        Pyro4.core.current_context.correlation_id = None


class ExposeDecoratorTests(unittest.TestCase):
    # note: the bulk of the tests for the @expose decorator are found in the test_util module
    def testExposeInstancemodeDefault(self):
        @Pyro4.core.expose
        class TestClassOne:
            def method(self):
                pass
        class TestClassTwo:
            @Pyro4.core.expose
            def method(self):
                pass
        class TestClassThree:
            def method(self):
                pass
        with Pyro4.core.Daemon() as daemon:
            daemon.register(TestClassOne)
            daemon.register(TestClassTwo)
            daemon.register(TestClassThree)
            self.assertEqual(("session", None), TestClassOne._pyroInstancing)
            self.assertEqual(("session", None), TestClassTwo._pyroInstancing)
            self.assertEqual(("session", None), TestClassThree._pyroInstancing)


class BehaviorDecoratorTests(unittest.TestCase):
    def testBehaviorInstancemodeInvalid(self):
        with self.assertRaises(ValueError):
            @Pyro4.core.behavior(instance_mode="kaputt")
            class TestClass:
                def method(self):
                    pass

    def testBehaviorRequiresParams(self):
        with self.assertRaises(SyntaxError) as x:
            @Pyro4.core.behavior
            class TestClass:
                def method(self):
                    pass
        self.assertIn("is missing argument", str(x.exception))

    def testBehaviorInstancecreatorInvalid(self):
        with self.assertRaises(TypeError):
            @Pyro4.core.behavior(instance_creator=12345)
            class TestClass:
                def method(self):
                    pass

    def testBehaviorOnMethodInvalid(self):
        with self.assertRaises(TypeError):
            class TestClass:
                @Pyro4.core.behavior(instance_mode="~invalidmode~")
                def method(self):
                    pass
        with self.assertRaises(TypeError):
            class TestClass:
                @Pyro4.core.behavior(instance_mode="percall", instance_creator=float)
                def method(self):
                    pass
        with self.assertRaises(TypeError):
            class TestClass:
                @Pyro4.core.behavior()
                def method(self):
                    pass

    def testBehaviorInstancing(self):
        @Pyro4.core.behavior(instance_mode="percall", instance_creator=float)
        class TestClass:
            def method(self):
                pass
        im, ic = TestClass._pyroInstancing
        self.assertEqual("percall", im)
        self.assertIs(float, ic)

    def testBehaviorWithExposeKeepsCorrectValues(self):
        @Pyro4.core.behavior(instance_mode="percall", instance_creator=float)
        @Pyro4.core.expose
        class TestClass:
            pass
        im, ic = TestClass._pyroInstancing
        self.assertEqual("percall", im)
        self.assertIs(float, ic)

        @Pyro4.core.expose
        @Pyro4.core.behavior(instance_mode="percall", instance_creator=float)
        class TestClass2:
            pass
        im, ic = TestClass2._pyroInstancing
        self.assertEqual("percall", im)
        self.assertIs(float, ic)


class RemoteMethodTests(unittest.TestCase):
    class BatchProxyMock(object):
        def __init__(self):
            self.result = []
            self._pyroMaxRetries = 0

        def __copy__(self):
            return self

        def __enter__(self):
            return self

        def __exit__(self, *args):
            pass

        def _pyroBatch(self):
            return Pyro4.core._BatchProxyAdapter(self)

        def _pyroInvokeBatch(self, calls, oneway=False):
            self.result = []
            for methodname, args, kwargs in calls:
                if methodname == "error":
                    self.result.append(Pyro4.futures._ExceptionWrapper(ValueError("some exception")))
                    break  # stop processing the rest, this is what Pyro should do in case of an error in a batch
                elif methodname == "pause":
                    time.sleep(args[0])
                self.result.append("INVOKED %s args=%s kwargs=%s" % (methodname, args, kwargs))
            if oneway:
                return
            else:
                return self.result

    class AsyncProxyMock(object):
        def __copy__(self):
            return self

        def __enter__(self):
            return self

        def __exit__(self, *args):
            pass

        def _pyroAsync(self, asynchronous=True):
            return self

        def __getattr__(self, item):
            return Pyro4.core._AsyncRemoteMethod(self, item, 5)

        def _pyroInvoke(self, methodname, vargs, kwargs, flags=0):
            if methodname == "pause_and_divide":
                time.sleep(vargs[0])
                return vargs[1] // vargs[2]
            else:
                raise NotImplementedError(methodname)

    def testRemoteMethodMetaOff(self):
        try:
            config.METADATA = False
            class ProxyMock(object):
                def invoke(self, name, args, kwargs):
                    return "INVOKED name=%s args=%s kwargs=%s" % (name, args, kwargs)

                def __getattr__(self, name):
                    return Pyro4.core._RemoteMethod(self.invoke, name, max_retries=0)

            o = ProxyMock()
            self.assertEqual("INVOKED name=foo args=(1,) kwargs={}", o.foo(1))  # normal
            self.assertEqual("INVOKED name=foo.bar args=(1,) kwargs={}", o.foo.bar(1))  # dotted
            self.assertEqual("INVOKED name=foo.bar args=(1, 'hello') kwargs={'a': True}", o.foo.bar(1, "hello", a=True))
            p = Pyro4.core.Proxy("PYRO:obj@host:666")
            a = p.someattribute
            self.assertIsInstance(a, Pyro4.core._RemoteMethod, "attribute access should just be a RemoteMethod")
            a2 = a.nestedattribute.nested2
            self.assertIsInstance(a2, Pyro4.core._RemoteMethod, "nested attribute should just be another RemoteMethod")
            p._pyroRelease()
        finally:
            config.METADATA = True

    def testRemoteMethodMetaOn(self):
        p = Pyro4.core.Proxy("PYRO:obj@localhost:59999")
        with self.assertRaises(Pyro4.errors.CommunicationError):
            _ = p.someattribute     # triggers attempt to get metadata
        p._pyroRelease()

    def testBatchMethod(self):
        proxy = self.BatchProxyMock()
        batch = Pyro4.core.batch(proxy)
        self.assertIsNone(batch.foo(42))
        self.assertIsNone(batch.bar("abc"))
        self.assertIsNone(batch.baz(42, "abc", arg=999))
        self.assertIsNone(batch.error())  # generate an exception
        self.assertIsNone(batch.foo(42))  # this call should not be performed after the error
        results = batch()
        result = next(results)
        self.assertEqual("INVOKED foo args=(42,) kwargs={}", result)
        result = next(results)
        self.assertEqual("INVOKED bar args=('abc',) kwargs={}", result)
        result = next(results)
        self.assertEqual("INVOKED baz args=(42, 'abc') kwargs={'arg': 999}", result)
        self.assertRaises(ValueError, next, results)  # the call to error() should generate an exception
        self.assertRaises(StopIteration, next, results)  # and now there should not be any more results
        self.assertEqual(4, len(proxy.result))  # should have done 4 calls, not 5
        batch._pyroRelease()

    def testBatchMethodOneway(self):
        proxy = self.BatchProxyMock()
        batch = Pyro4.core.batch(proxy)
        self.assertIsNone(batch.foo(42))
        self.assertIsNone(batch.bar("abc"))
        self.assertIsNone(batch.baz(42, "abc", arg=999))
        self.assertIsNone(batch.error())  # generate an exception
        self.assertIsNone(batch.foo(42))  # this call should not be performed after the error
        results = batch(oneway=True)
        self.assertIsNone(results)  # oneway always returns None
        self.assertEqual(4, len(proxy.result))  # should have done 4 calls, not 5
        self.assertRaises(Pyro4.errors.PyroError, batch, oneway=True, asynchronous=True)  # oneway+asynchronous=booboo

    def testBatchMethodAsync(self):
        proxy = self.BatchProxyMock()
        batch = Pyro4.core.batch(proxy)
        self.assertIsNone(batch.foo(42))
        self.assertIsNone(batch.bar("abc"))
        self.assertIsNone(batch.pause(0.5))  # pause shouldn't matter with asynchronous
        self.assertIsNone(batch.baz(42, "abc", arg=999))
        begin = time.time()
        asyncresult = batch(asynchronous=True)
        duration = time.time() - begin
        self.assertLess(duration, 0.2, "batch oneway with pause should still return almost immediately")
        results = asyncresult.value
        self.assertEqual(4, len(proxy.result))  # should have done 4 calls
        result = next(results)
        self.assertEqual("INVOKED foo args=(42,) kwargs={}", result)
        result = next(results)
        self.assertEqual("INVOKED bar args=('abc',) kwargs={}", result)
        result = next(results)
        self.assertEqual("INVOKED pause args=(0.5,) kwargs={}", result)
        result = next(results)
        self.assertEqual("INVOKED baz args=(42, 'abc') kwargs={'arg': 999}", result)
        self.assertRaises(StopIteration, next, results)  # and now there should not be any more results

    def testBatchMethodReuse(self):
        proxy = self.BatchProxyMock()
        batch = Pyro4.core.batch(proxy)
        batch.foo(1)
        batch.foo(2)
        results = batch()
        self.assertEqual(['INVOKED foo args=(1,) kwargs={}', 'INVOKED foo args=(2,) kwargs={}'], list(results))
        # re-use the batch proxy:
        batch.foo(3)
        batch.foo(4)
        results = batch()
        self.assertEqual(['INVOKED foo args=(3,) kwargs={}', 'INVOKED foo args=(4,) kwargs={}'], list(results))
        results = batch()
        self.assertEqual(0, len(list(results)))

    def testChangedAsyncLegacyApi(self):
        try:
            Pyro4.config.METADATA = False
            proxy = Pyro4.Proxy("PYRO:test@addr:5555")
            self.assertIsNone(proxy._pyroAsync())
            m = proxy.foo
            self.assertIsInstance(m, Pyro4.core._AsyncRemoteMethod)
            self.assertIsNone(proxy._pyroAsync(False))
            m = proxy.foo
            self.assertIsInstance(m, Pyro4.core._RemoteMethod)
            self.assertIsNone(Pyro4.asyncproxy(proxy))
            m = proxy.foo
            self.assertIsInstance(m, Pyro4.core._AsyncRemoteMethod)
            self.assertIsNone(Pyro4.asyncproxy(proxy, False))
            m = proxy.foo
            self.assertIsInstance(m, Pyro4.core._RemoteMethod)
        finally:
            Pyro4.config.METADATA = True

    def testAsyncMethod(self):
        proxy = self.AsyncProxyMock()
        Pyro4.core.asyncproxy(proxy)
        begin = time.time()
        result = proxy.pause_and_divide(0.2, 10, 2)  # returns immediately
        duration = time.time() - begin
        self.assertLess(duration, 0.1)
        self.assertFalse(result.ready)
        _ = result.value
        self.assertTrue(result.ready)
        proxy._pyroRelease()

    def testAsyncCallbackMethod(self):
        class AsyncFunctionHolder(object):
            asyncFunctionCount = 0

            def asyncFunction(self, value, amount=1):
                self.asyncFunctionCount += 1
                return value + amount

        proxy = self.AsyncProxyMock()
        Pyro4.core.asyncproxy(proxy)
        result = proxy.pause_and_divide(0.2, 10, 2)  # returns immediately
        holder = AsyncFunctionHolder()
        result.then(holder.asyncFunction, amount=2) \
            .then(holder.asyncFunction, amount=4) \
            .then(holder.asyncFunction)
        value = result.value
        self.assertEqual(10 // 2 + 2 + 4 + 1, value)
        self.assertEqual(3, holder.asyncFunctionCount)
        proxy._pyroRelease()

    def testCrashingAsyncCallbackMethod(self):
        def normalAsyncFunction(value, x):
            return value + x

        def crashingAsyncFunction(value):
            return 1 // 0  # crash

        proxy = self.AsyncProxyMock()
        Pyro4.core.asyncproxy(proxy)
        result = proxy.pause_and_divide(0.2, 10, 2)  # returns immediately
        result.then(crashingAsyncFunction).then(normalAsyncFunction, 2)
        try:
            _ = result.value
            self.fail("expected exception")
        except ZeroDivisionError:
            pass  # ok
        proxy._pyroRelease()

    def testAsyncMethodTimeout(self):
        proxy = self.AsyncProxyMock()
        Pyro4.core.asyncproxy(proxy)
        result = proxy.pause_and_divide(1, 10, 2)  # returns immediately
        self.assertFalse(result.ready)
        self.assertFalse(result.wait(0.5))  # won't be ready after 0.5 sec
        self.assertTrue(result.wait(1))  # will be ready within 1 seconds more
        self.assertTrue(result.ready)
        self.assertEqual(5, result.value)
        proxy._pyroRelease()


class TestSimpleServe(unittest.TestCase):
    class DaemonWrapper(Pyro4.core.Daemon):
        def requestLoop(self, *args):
            # override with empty method to fall out of the serveSimple call
            pass

    def testSimpleServe(self):
        with TestSimpleServe.DaemonWrapper() as d:
            o1 = MyThingPartlyExposed(1)
            o2 = MyThingPartlyExposed(2)
            objects = {o1: "test.o1", o2: None}
            Pyro4.core.Daemon.serveSimple(objects, daemon=d, ns=False, verbose=False)
            self.assertEqual(3, len(d.objectsById))
            self.assertIn("test.o1", d.objectsById)
            self.assertIn(o1, d.objectsById.values())
            self.assertIn(o2, d.objectsById.values())

    def testSimpleServeSameNames(self):
        with TestSimpleServe.DaemonWrapper() as d:
            o1 = MyThingPartlyExposed(1)
            o2 = MyThingPartlyExposed(2)
            o3 = MyThingPartlyExposed(3)
            objects = {o1: "test.name", o2: "test.name", o3: "test.othername"}
            with self.assertRaises(Pyro4.errors.DaemonError):
                Pyro4.core.Daemon.serveSimple(objects, daemon=d, ns=False, verbose=False)


def futurestestfunc(a, b, extra=None):
    if extra is None:
        return a + b
    else:
        return a + b + extra


def crashingfuturestestfunc(a):
    return 1 // 0  # crash


class FuturesErrorHandlerStorage(object):
    def __init__(self):
        self.error = None
    def errorhandler(self, err):
        self.error = err


class TestFutures(unittest.TestCase):
    def testSimpleFuture(self):
        f = Pyro4.futures.Future(futurestestfunc)
        r = f(4, 5)
        self.assertIsInstance(r, Pyro4.futures.FutureResult)
        value = r.value
        self.assertEqual(9, value)

    def testFutureChain(self):
        f = Pyro4.futures.Future(futurestestfunc) \
            .then(futurestestfunc, 6) \
            .then(futurestestfunc, 7) \
            .then(futurestestfunc, 8) \
            .then(futurestestfunc, 9, extra=10)
        r = f(4, 5)
        value = r.value
        self.assertEqual(4 + 5 + 6 + 7 + 8 + 9 + 10, value)

    def testCrashingChain(self):
        f = Pyro4.futures.Future(futurestestfunc) \
            .then(futurestestfunc, 6) \
            .then(crashingfuturestestfunc) \
            .then(futurestestfunc, 8)
        r = f(4, 5)
        try:
            _ = r.value
            self.fail("expected exception")
        except ZeroDivisionError:
            pass  # ok

    def testErrorHandler(self):
        storage = FuturesErrorHandlerStorage()
        f = Pyro4.futures.Future(crashingfuturestestfunc) \
            .then(futurestestfunc, 5) \
            .iferror(storage.errorhandler) \
            .then(futurestestfunc, 6)
        self.assertIsNone(storage.error)
        r = f(42)
        try:
            _ = r.value
        except ZeroDivisionError:
            pass  # ok
        self.assertIsInstance(storage.error, ZeroDivisionError)

    def testFutureResultChainSlow(self):
        f = Pyro4.futures.Future(futurestestfunc)
        result = f(4, 5)
        time.sleep(.02)
        result.then(futurestestfunc, 6)
        time.sleep(.02)
        result.then(futurestestfunc, 7)
        time.sleep(.02)
        result.then(futurestestfunc, 8)
        time.sleep(.02)
        result.then(futurestestfunc, 9)
        time.sleep(.02)
        result.then(futurestestfunc, 10)
        time.sleep(.02)
        value = result.value
        self.assertEqual(4 + 5 + 6 + 7 + 8 + 9 + 10, value)

    def testFutureResultChain(self):
        f = Pyro4.futures.Future(futurestestfunc)
        result = f(4, 5).then(futurestestfunc, 6).then(futurestestfunc, 7).then(futurestestfunc, 8).then(futurestestfunc, 9).then(futurestestfunc, 10)
        value = result.value
        self.assertEqual(4 + 5 + 6 + 7 + 8 + 9 + 10, value)
        with self.assertRaises(RuntimeError):
            f(4, 5)   # cannot evaluate the same future more than once

    def testFutureDelay(self):
        f = Pyro4.futures.Future(futurestestfunc)
        b = f.delay(0)
        self.assertTrue(b)
        begin = time.time()
        f(4, 5).value
        duration = time.time() - begin
        self.assertLess(duration, 0.1)
        f = Pyro4.futures.Future(futurestestfunc)
        b = f.delay(1)
        self.assertTrue(b)
        begin = time.time()
        r = f(4, 5)
        duration = time.time() - begin
        self.assertLess(duration, 0.1)
        begin = time.time()
        r.value
        duration = time.time() - begin
        self.assertGreaterEqual(duration, 0.98)
        self.assertLess(duration, 1.1)
        self.assertFalse(f.delay(10))

    def testFutureCancel(self):
        f = Pyro4.futures.Future(futurestestfunc)
        f.delay(10)
        b = f.cancel()
        self.assertTrue(b)
        with self.assertRaises(RuntimeError) as x:
            f(4, 5)
        self.assertTrue("cancelled" in str(x.exception))
        f = Pyro4.futures.Future(futurestestfunc)
        f.delay(10)
        result = f(4, 5)
        b = f.cancel()
        self.assertTrue(b)
        success = result.wait(3)
        self.assertTrue(success)
        with self.assertRaises(RuntimeError) as x:
            result.value
        self.assertTrue("cancelled" in str(x.exception))
        f = Pyro4.futures.Future(futurestestfunc)
        result = f(4, 5)
        result.value
        b = f.cancel()
        self.assertFalse(b)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



from __future__ import print_function

from awesome_thirdparty_library import AwesomeClass
import Pyro4


# expose the class from the library using @expose as wrapper function:
ExposedClass = Pyro4.expose(AwesomeClass)


with Pyro4.Daemon() as daemon:
    # register the wrapped class instead of the library class itself:
    uri = daemon.register(ExposedClass, "example.thirdpartylib")
    print("wrapped class registered, uri: ", uri)
    daemon.requestLoop()



from __future__ import print_function
import Pyro4.core
import Pyro4.naming
import chain

this_node = "C"
next_node = "A"

servername = "example.chain." + this_node

with Pyro4.core.Daemon() as daemon:
    obj = chain.Chain(this_node, next_node)
    uri = daemon.register(obj)
    with Pyro4.naming.locateNS() as ns:
        ns.register(servername, uri)

    # enter the service loop.
    print("Server started %s" % this_node)
    daemon.requestLoop()



import random


class StockMarket(object):
    def __init__(self, marketname, symbols):
        self.name = marketname
        self.symbolmeans = {}
        for symbol in symbols:
            self.symbolmeans[symbol] = random.uniform(20, 200)
        self.aggregators = []

    def generate(self):
        quotes = {}
        for symbol, mean in self.symbolmeans.items():
            if random.random() < 0.2:
                quotes[symbol] = round(random.normalvariate(mean, 20), 2)
        for aggregator in self.aggregators:
            aggregator.quotes(self.name, quotes)

    def listener(self, aggregator):
        self.aggregators.append(aggregator)

    def symbols(self):
        return self.symbolmeans.keys()



from __future__ import print_function
import datetime
import pytz
import dateutil.tz
import Pyro4


fmt = '%Y-%m-%d %H:%M:%S %Z%z'

Pyro4.config.SERIALIZERS_ACCEPTED = {"pickle", "marshal", "json", "serpent"}


@Pyro4.expose
class Server(object):
    def echo(self, date):
        print("ECHO:")
        print(" [raw] ", repr(date))
        if hasattr(date, "isoformat"):
            print(" [iso] ", date.isoformat())
        return date

    def pytz(self):
        tz_nl = pytz.timezone("Europe/Amsterdam")
        return tz_nl.localize(datetime.datetime.now())

    def dateutil(self):
        tz_nl = dateutil.tz.gettz("Europe/Amsterdam")
        return datetime.datetime.now(tz_nl)

# main program

Pyro4.Daemon.serveSimple({
    Server: "example.timezones"
}, ns=False)



from __future__ import print_function
import sys
import random
import Pyro4


if sys.version_info < (3, 0):
    input = raw_input


uri = input("Enter the URI of the server object: ")

with Pyro4.Proxy(uri) as proxy:
    print("currently allocated resources:", proxy.list())
    name1 = hex(random.randint(0, 999999))[-4:]
    name2 = hex(random.randint(0, 999999))[-4:]
    print("allocating resource...", name1)
    proxy.allocate(name1)
    print("allocating resource...", name2)
    proxy.allocate(name2)
    input("\nhit Enter now to continue normally or ^C/break to abort the connection forcefully:")
    print("free resources normally...")
    proxy.free(name1)
    proxy.free(name2)
    print("allocated resources:", proxy.list())


print("done.")



from __future__ import print_function
import sys
import warnings

import Pyro4
import Pyro4.message


warnings.filterwarnings("ignore")

if sys.version_info < (3, 0):
    input = raw_input

print("You can run this client on a different computer so you can disable the network connection (by yanking out the lan cable or whatever).")
print("Alternatively, wait for a timeout on the server, which will then close its connection.")
uri = input("Uri of server? ").strip()


class AutoReconnectingProxy(Pyro4.core.Proxy):
    """
    A Pyro proxy that automatically recovers from a server disconnect.
    It does this by intercepting every method call and then it first 'pings'
    the server to see if it still has a working connection. If not, it
    reconnects the proxy and retries the method call.
    Drawback is that every method call now uses two remote messages (a ping,
    and the actual method call).
    This uses some advanced features of the Pyro API.
    """

    def _pyroInvoke(self, *args, **kwargs):
        # We override the method that does the actual remote calls: _pyroInvoke.
        # If there's still a connection, try a ping to see if it is still alive.
        # If it isn't alive, reconnect it. If there's no connection, simply call
        # the original method (it will reconnect automatically).
        if self._pyroConnection:
            try:
                print("  <proxy: ping>")
                Pyro4.message.Message.ping(self._pyroConnection, hmac_key=None)    # utility method on the Message class
                print("  <proxy: ping reply (still connected)>")
            except Pyro4.errors.ConnectionClosedError:
                print("  <proxy: Connection lost. REBINDING...>")
                self._pyroReconnect()
                print("  <proxy: Connection restored, continue with actual method call...>")
        return super(AutoReconnectingProxy, self)._pyroInvoke(*args, **kwargs)


with AutoReconnectingProxy(uri) as obj:
    result = obj.echo("12345")
    print("result =", result)
    print("\nClient proxy connection is still open. Disable the network now (or wait until the connection timeout on the server expires) and see what the server does.")
    print("Once you see on the server that it got a timeout or a disconnect, enable the network again.")
    input("Press enter to continue:")
    print("\nDoing a new call on the same proxy:")
    result = obj.echo("12345")
    print("result =", result)



# This is the code that runs this example.
from warehouse import Warehouse
from person import Person

warehouse = Warehouse()
janet = Person("Janet")
henry = Person("Henry")
janet.visit(warehouse)
henry.visit(warehouse)



from __future__ import print_function, division
import sys
import Pyro4
import Pyro4.util

if sys.version_info < (3, 0):
    input = raw_input

sys.excepthook = Pyro4.util.excepthook


uri = input("Enter streaming server uri: ").strip()
with Pyro4.Proxy(uri) as p:
    print("\nnormal list:")
    print(p.list())
    print("\nvia iterator:")
    print(list(p.iterator()))
    print("\nvia generator:")
    print(list(p.generator()))
    print("\nslow generator:")
    for number in p.slow_generator():
        print(number)



from __future__ import print_function
import time

import Pyro4
from shoppingcart import ShoppingCart


@Pyro4.expose
@Pyro4.behavior(instance_mode="single")
class Shop(object):
    inventory = {
        "paper": 1.25,
        "bread": 1.50,
        "meat": 5.99,
        "milk": 0.80,
        "fruit": 2.65,
        "chocolate": 3.99,
        "pasta": 0.50,
        "sauce": 1.20,
        "vegetables": 1.40,
        "cookies": 1.99,
        "pizza": 3.60,
        "shampoo": 2.22,
        "whiskey": 24.99
    }

    customersInStore = {}

    def enter(self, name):
        print("Customer %s enters the store." % name)
        print("Customer takes a shopping cart.")
        # create a cart and return it as a pyro object to the client
        cart = ShoppingCart()
        self.customersInStore[name] = cart
        self._pyroDaemon.register(cart)  # make cart a pyro object
        return cart

    def customers(self):
        return list(self.customersInStore.keys())

    def goods(self):
        return self.inventory

    def payByName(self, name):
        print("Customer %s goes to the counter to pay." % name)
        cart = self.customersInStore[name]
        return self.payCart(cart, name)

    def payCart(self, cart, name=None):
        receipt = []
        if name:
            receipt.append("Receipt for %s." % name)
        receipt.append("Receipt Date: " + time.asctime())
        total = 0.0
        for item in cart.getContents():
            price = self.inventory[item]
            total += price
            receipt.append("%13s  %.2f" % (item, price))
        receipt.append("")
        receipt.append("%13s  %.2f" % ("total:", total))
        cart.empty()
        return "\n".join(receipt)

    def leave(self, name):
        print("Customer %s leaves." % name)
        cart = self.customersInStore[name]
        print("  their shopping cart contains: %s" % cart.getContents())
        if cart.getContents():
            print("  it is not empty, they are trying to shoplift!")
            raise Exception("attempt to steal a full cart prevented")
        # delete the cart and unregister it with pyro
        del self.customersInStore[name]
        self._pyroDaemon.unregister(cart)


# main program

Pyro4.Daemon.serveSimple({
    Shop: "example.shop"
})
